
#include <assert.h>
#include <math.h>
#include "global.h"
#include "processor.h"
#include "list.h"
#include "dag.h"
#include "csp.h"
#include "timer.h"


/* Variables for CSP-specific settings */
double timeoutLevels[NUM_TIMEOUT_LEVELS];
int    maxRegionSize, maxRestartSize;


/* =================================================================
 *  Routines for creating and destroying a CSP with n variables.
 */

/*
 *  Allocate space for n variables.
 */
CSP_type *
createCSP( int n )
{
    int       i;
    CSP_type  *C;

    C = (CSP_type *)calloc( (size_t)1, sizeof(CSP_type) );

    C->n = n;
    C->visits = 0;
    C->backtracks = 0;

    C->varType = (int *)calloc( (size_t)n+1, sizeof(int) );
#ifdef USE_PIPELINING
	C->varDuration = (int *)calloc( (size_t)n+1, sizeof(int) );
	C->nDummyNodes = 0;
#endif

#ifdef USE_EXECUTION_SERIALIZATION
	C->varIsSerial = (BOOL *)calloc( (size_t)n+1, sizeof(int) );
#endif

    C->nOps    = (int *)calloc( (size_t)N_INSTRUCTION_TYPES+1, sizeof(int) );
    C->lower   = (int *)calloc( (size_t)n+1, sizeof(int) );
    C->upper   = (int *)calloc( (size_t)n+1, sizeof(int) );

    C->nSideExits = 0;
    C->costCoeffs = (int *)calloc( (size_t)n, sizeof(int) );
    C->costFreqs  = (double *)calloc( (size_t)n, sizeof(double) );
    C->costVars   = (int *)calloc( (size_t)n, sizeof(int) );

    C->cp = (int **)calloc( (size_t)C->n+1, sizeof(int *) );
    for( i = 1; i <= C->n; i++ ) {
	C->cp[i] = (int *)calloc( (size_t)C->n+1, sizeof(int) );
    }

    C->fixed = (BOOL *)calloc( (size_t)n+1, sizeof(BOOL) );
    C->isArticulationNode = (BOOL *)calloc( (size_t)n+1, sizeof(BOOL) );

    C->edgeMarked = (BOOL **)calloc( (size_t)C->n+1, sizeof(BOOL *) );
    for( i = 1; i <= C->n; i++ ) {
	C->edgeMarked[i] = (BOOL *)calloc( (size_t)C->n+1, sizeof(BOOL) );
    }

    C->latencySucc = (ELEM **) calloc( (size_t)n+1, sizeof(ELEM *) );
    C->latencyPred = (ELEM **) calloc( (size_t)n+1, sizeof(ELEM *) );
    for( i = 1; i <= n; i++ ) {
	C->latencySucc[i] = (ELEM *)NULL;
	C->latencyPred[i] = (ELEM *)NULL;
    }

    C->distanceSucc = (ELEM **) calloc( (size_t)n+1, sizeof(ELEM *) );
    C->distancePred = (ELEM **) calloc( (size_t)n+1, sizeof(ELEM *) );
    for( i = 1; i <= n; i++ ) {
	C->distanceSucc[i] = (ELEM *)NULL;
	C->distancePred[i] = (ELEM *)NULL;
    }

    C->distanceUpperSucc = (ELEM **) calloc( (size_t)n+1, sizeof(ELEM *) );
    C->distanceUpperPred = (ELEM **) calloc( (size_t)n+1, sizeof(ELEM *) );
    for( i = 1; i <= n; i++ ) {
	C->distanceUpperSucc[i] = (ELEM *)NULL;
	C->distanceUpperPred[i] = (ELEM *)NULL;
    }

    C->dominanceSucc = (ELEM **) calloc( (size_t)n+1, sizeof(ELEM *) );
    C->dominancePred = (ELEM **) calloc( (size_t)n+1, sizeof(ELEM *) );
    for( i = 1; i <= n; i++ ) {
	C->dominanceSucc[i] = (ELEM *)NULL;
	C->dominancePred[i] = (ELEM *)NULL;
    }

    C->edgeSucc = (ELEM **) calloc( (size_t)n+1, sizeof(ELEM *) );
    C->edgePred = (ELEM **) calloc( (size_t)n+1, sizeof(ELEM *) );
    for( i = 1; i <= n; i++ ) {
	C->edgeSucc[i] = (ELEM *)NULL;
	C->edgePred[i] = (ELEM *)NULL;
    }

    C->nPreds  = (int *)calloc( (size_t)n+1, sizeof(int) );
    C->nSuccs  = (int *)calloc( (size_t)n+1, sizeof(int) );
    C->changev = (int *)calloc( (size_t)n+1, sizeof(int) );

    C->saveLower = (int **)calloc( (size_t)C->n+1, sizeof(int *) );
    C->saveUpper = (int **)calloc( (size_t)C->n+1, sizeof(int *) );
    for( i = 0; i <= C->n; i++ ) {
	C->saveLower[i] = (int *)calloc( (size_t)C->n+1, sizeof(int) );
	C->saveUpper[i] = (int *)calloc( (size_t)C->n+1, sizeof(int) );
    }

    C->s_lower = createStack( n+1 );
    C->s_upper = createStack( n+1 );

    C->values = (node_type *)calloc( (size_t)n+1, sizeof(node_type) );

    C->filled = (BOOL *)calloc( (size_t)getMaxLatency()*(n+1), sizeof(BOOL) );

    C->iv        = (interval  *)calloc((size_t)n, sizeof(interval  ));
    C->minsorted = (interval **)calloc((size_t)n, sizeof(interval *));
    C->maxsorted = (interval **)calloc((size_t)n, sizeof(interval *));
    C->bounds    = (int *)calloc((size_t)(2*n+2), sizeof(int));

    C->t = (int *)calloc((size_t)(2*n+2), sizeof(int));
    C->d = (int *)calloc((size_t)(2*n+2), sizeof(int));
    C->h = (int *)calloc((size_t)(2*n+2), sizeof(int));

    set_start_time_CSP( C );

    return( C );
}

void
destroyCSP( CSP_type *C )
{
    int  i;

    free( C->varType );
#ifdef USE_PIPELINING
	free( C->varDuration );
#endif

#ifdef USE_EXECUTION_SERIALIZATION
	free( C->varIsSerial );
#endif
    free( C->nOps );
    free( C->lower );
    free( C->upper );
    free( C->costCoeffs );
    free( C->costFreqs );
    free( C->costVars );

    for( i = 1; i <= C->n; i++ ) {
	free( C->cp[i] );
    }
    free( C->cp );

    free( C->fixed );
    free( C->isArticulationNode );

    for( i = 1; i <= C->n; i++ ) {
	free( C->edgeMarked[i] );
    }
    free( C->edgeMarked );

    for( i = 1; i <= C->n; i++ ) {
	destroyList( C->latencySucc[i] );
	destroyList( C->latencyPred[i] );
    }
    free( C->latencySucc );
    free( C->latencyPred );

    for( i = 1; i <= C->n; i++ ) {
	destroyList( C->distanceSucc[i] );
	destroyList( C->distancePred[i] );
    }
    free( C->distanceSucc );
    free( C->distancePred );

    for( i = 1; i <= C->n; i++ ) {
	destroyList( C->distanceUpperSucc[i] );
	destroyList( C->distanceUpperPred[i] );
    }
    free( C->distanceUpperSucc );
    free( C->distanceUpperPred );

    for( i = 1; i <= C->n; i++ ) {
	destroyList( C->dominanceSucc[i] );
	destroyList( C->dominancePred[i] );
    }
    free( C->dominanceSucc );
    free( C->dominancePred );

    for( i = 1; i <= C->n; i++ ) {
	destroyList( C->edgeSucc[i] );
	destroyList( C->edgePred[i] );
    }
    free( C->edgeSucc );
    free( C->edgePred );

    free( C->nPreds );
    free( C->nSuccs );
    free( C->changev );

    for( i = 0; i <= C->n; i++ ) {
	free( C->saveLower[i] );
	free( C->saveUpper[i] );
    }
    free( C->saveLower );
    free( C->saveUpper );

    destroyStack( C->s_lower );
    destroyStack( C->s_upper );

    free( C->values );

    free( C->filled );

    free( C->iv );
    free( C->minsorted );
    free( C->maxsorted );
    free( C->bounds );

    free( C->h );
    free( C->d );
    free( C->t );

    free( C );
}


/* =================================================================
 *  Routines for the cost function.
 */

/*
 *  Set the cost function of a CSP.
 *
 *  In the DAG, side exits are labeled with the probability
 *  (represented as a percentage = 1, 10, 20, ..., 90) that the
 *  side exit will be taken, given that the flow of control
 *  has reached that point. This is converted to an overall
 *  probability (represented as a percentage = 1, 2, ..., 99)
 *  that the flow of control will exit the basic block at this
 *  branch point.
 */
void
setCostFunction( CSP_type *C, DAG_type *DAG )
{
    int    i;
    double freq_fallthrough;
    double percent, coeff;

    /*
     *  Determine the probability of exiting at each side exit.
     */
    percent = 1.0;
    C->nSideExits = 0;
    // -1 since the last node cannot be a side exit.
    for( i = 0; i < DAG->nInstructions-1; i++ ) {
        if( DAG->percentage[i] ) {
            coeff = percent * ((double)DAG->percentage[i]);
#ifdef WIN32
	    // No rounding function on Win32 in the math library
	    C->costCoeffs[C->nSideExits] = (int)coeff;
#else
            C->costCoeffs[C->nSideExits] = (int)rint(coeff);
#endif
	    C->costFreqs[C->nSideExits] = DAG->f1[i] + DAG->f2[i];
	    if( C->costFreqs[C->nSideExits] == 0.0 ) {
	        C->costFreqs[C->nSideExits] = 1.0;
	    }
            C->costVars[C->nSideExits] = i+1;
            if( C->costCoeffs[C->nSideExits] > 0 ) {
		C->nSideExits++;
	    }
            percent *= (1.0 - ((double)DAG->percentage[i])/100.0);
        }
    }
    /*
     *  Determine the probability that we fall through to the sink node.
     */
    C->costCoeffs[C->nSideExits] = 100;
    for( i = 0; i < C->nSideExits; i++ ) {
        C->costCoeffs[C->nSideExits] -= C->costCoeffs[i];
    }
    freq_fallthrough = 1.0E100;
    for( i = 0; i < DAG->nInstructions-1; i++ ) {
        if( DAG->percentage[i] && DAG->f2[i] > 0.0 ) {
	    if( freq_fallthrough > DAG->f2[i] ) {
		freq_fallthrough = DAG->f2[i];
	    }
	}
    }
    if( freq_fallthrough == 0.0 || freq_fallthrough == 1.0E100 ) {
	freq_fallthrough = 1.0;
    }
    C->costFreqs[C->nSideExits] = freq_fallthrough;
    /*
     *  Rarely the fall through probability is zero or negative
     *  (if rounding goes against us). If so, set it to the
     *  smallest value.
     */
    if( C->costCoeffs[C->nSideExits] <= 0 ) {
        C->costCoeffs[C->nSideExits] = 1;
    }
    C->costVars[C->nSideExits] = C->n;
}

/*
 *  Print the cost function of a CSP.
 */
void
printCostFunction( CSP_type *C )
{
    int i;

    printf("\n\n");
    for( i = 0; i < C->nSideExits; i++ ) {
	printf("%d x_%d + ", C->costCoeffs[i], C->costVars[i]);
    }
    printf("%d x_%d\n",
	C->costCoeffs[C->nSideExits], C->costVars[C->nSideExits]);
    for( i = 0; i < C->nSideExits; i++ ) {
	printf("%0.0f ", C->costFreqs[i]);
    }
    printf("%0.0f\n",
	C->costFreqs[C->nSideExits]);
}

/*
 *  Get the value of the cost function of a CSP.
 *
 *  If called with a solution
 *  (i.e., C->lower[i] == C->upper[i], for all i),
 *  then the value returned is the cost of the solution.
 *
 *  If called before a solution has been found
 *  (i.e., C->lower[i] <= C->upper[i], for all i),
 *  then the value returned is a lower bound on the cost of a solution.
 */
int
getCost( CSP_type *C )
{
    int i, cost;

    cost = 0;
    for( i = 0; i <= C->nSideExits; i++ ) {
	cost += C->costCoeffs[i] * C->lower[C->costVars[i]];
    }

    return( cost );
}

/*
 *  Get the cost of a schedule.
 */
int
getCostSchedule( CSP_type *C, int schedule[] )
{
    int i, cost;

    cost = 0;
    for( i = 0; i <= C->nSideExits; i++ ) {
	cost += C->costCoeffs[i] * schedule[C->costVars[i]-1];
    }

    return( cost );
}

/*
 *  Given a lower bound on the cost function and upper bounds
 *  on the values of the variables in the cost function, determine
 *  a lower bound on the j'th cost variable. Can also be used to
 *  find a lower bound on the length of an optimal schedule if
 *  j is equal to the final exit (the sink node).
 */
int
getLengthLowerBound( CSP_type *C, int j )
{
    int i, cost;

    cost = C->lbCost;
    for( i = 0; i <= C->nSideExits; i++ ) if( i != j ) {
	cost -= C->costCoeffs[i] * C->upper[C->costVars[i]];
    }

    /*
     *  Want the ceiling of this value
     */
    return( (int)ceil((double)cost / (double)C->costCoeffs[j]) );
}

/*
 *  Given an upper bound on the cost function and lower bounds
 *  on the values of the variables in the cost function, determine
 *  an upper bound on the j'th cost variable. Can also be used to
 *  find an upper bound on the length of an optimal schedule if
 *  j is equal to the final exit (the sink node).
 */
int
getLengthUpperBound( CSP_type *C, int j )
{
    int i, cost;

    cost = C->ubCost;
    for( i = 0; i <= C->nSideExits; i++ ) if( i != j ) {
	cost -= C->costCoeffs[i] * C->lower[C->costVars[i]];
    }

    /*
     *  Want the floor of this value, so use integer division.
     */
    return( cost / C->costCoeffs[j] );
}

void
enumerateCostVariables( CSP_type *C, int **all, int *n_assignments, int *partial, int current )
{
    int i, a, b, cost, consistent, completionCost;

    saveBounds( C, current+1 );

    if( (*n_assignments) >= MAX_NUMBER_ASSIGNS ) {
	// Too many possibilties...
	return;
    }

    if( current > C->nSideExits ) {
	cost = 0;
	for( i = 0; i <= C->nSideExits; i++ ) {
	    cost += C->costCoeffs[i] * partial[i];
	}
	if( (C->lbCost <= cost) && (cost <= C->ubCost) ) {
	    all[*n_assignments] = (int *)calloc( (size_t)C->nSideExits+1, sizeof(int) );
	    for( i = 0; i <= C->nSideExits; i++ ) {
	        all[*n_assignments][i] = partial[i];
	    }
	    (*n_assignments)++;
	}
	return;
    }

    a = C->lower[C->costVars[current]];
    b = C->upper[C->costVars[current]];

    for( ; a <= b; a++ ) {

	partial[current] = a;

	C->lower[C->costVars[current]] = a;
	C->upper[C->costVars[current]] = a;

	consistent = propagate( C, C->costVars[current], current );
	if( !consistent ) {
            restoreBounds( C, current+1 );
	    continue;
	}

	cost = 0;
	for( i = 0; i <= current; i++ ) {
	    cost += C->costCoeffs[i] * partial[i];
	}
	/*
	 *  If cost of partial assignment + lower bound on completion > ubCost
	 *    return;
	 */
	completionCost = 0;
	for( i = current+1; i <= C->nSideExits; i++ ) {
	    completionCost += C->costCoeffs[i] * C->lower[C->costVars[i]];
	}
	if( (cost + completionCost) > C->ubCost ) {
            restoreBounds( C, current+1 );
	    continue;
	}
	/*
	 *  If cost of partial assignment + upper bound on completion < lbCost
	 *    return;
	 */
	completionCost = 0;
	for( i = current+1; i <= C->nSideExits; i++ ) {
	    completionCost += C->costCoeffs[i] * C->upper[C->costVars[i]];
	}
	if( (cost + completionCost) < C->lbCost ) {
            restoreBounds( C, current+1 );
	    continue;
	}

	enumerateCostVariables( C, all, n_assignments, partial, current+1 );

        restoreBounds( C, current+1 );
    }
}

/*
 *  Determine whether all the side exits are articulation nodes.
 *  Precondition: critical path distances, cp[i][j], must be set.
 */
void
setArticulationNodes( CSP_type *C )
{

    int  w, latest;
    ELEM *p;

    for( w = 1; w <= C->n; w++ ) {
	C->isArticulationNode[w] = FALSE;
    }
    C->isArticulationNode[1] = TRUE;

    /*
     *  Mark all articulation nodes between 1 and the sink node (C->n).
     */
    latest = 1;
    while( latest < C->n ) {
	latest++;
        for( w = 1; w <= C->n; w++ ) {
            if( w == latest ) {
                break;
            }
            for( p = C->latencySucc[w]; p != NULL; p = p->next ) {
                if( p->vertex > latest ) {
                    latest = p->vertex;
                }
            }
        }
	C->isArticulationNode[latest] = TRUE;
	//printf("x_%d is an articulation node\n",latest);
    }
}

/*
 *  Test whether all of the cost variables are articulation nodes.
 *  Precondition: setArticulationNodes has been called
 */
int
allArticulationNodes( CSP_type *C )
{
    int  i;

    for( i = 0; i < C->nSideExits; i++ ) {
	if( !(C->isArticulationNode[C->costVars[i]]) ) {
	    return( FALSE );
	}
    }

    return( TRUE );
}

/* =================================================================
 *  Routines for manipulating the bounds of the CSP variables.
 */

static void
topologicalSort(
	CSP_type *C,
	int v,
	int *order,
	int *index,
	BOOL *marked )
{
    int   w;
    ELEM  *p;

    /*
     *  For each successor w of v that is not already marked...
     */
    for( p = C->latencySucc[v]; p != NULL; p = p->next ) {
	w = p->vertex;
	if( !marked[w] ) {
	    topologicalSort( C, w, order, index, marked );
	}
    }
    for( p = C->distanceSucc[v]; p != NULL; p = p->next ) {
	w = p->vertex;
	if( !marked[w] ) {
	    topologicalSort( C, w, order, index, marked );
	}
    }

    /*
     *  Insert the node into the order.
     */
    marked[v] = TRUE;
    order[*index] = v;
    (*index)--;
}


/*
 *  Set and improve the lower bounds of the variables by
 *  propagating the latency, distance, and edge constraints.
 */
static void
setLowerBounds( CSP_type *C, int order[] )
{
    int  i, v, w, type;
    ELEM *p;

    for( v = 1; v <= C->n; v++ ) {
	C->lower[v] = 1;
    }

    for( i = 1; i <= C->n; i++ ) {
	v = order[i];

	if( C->nPreds[v] > 1 ) {
	    for( type = 0; type < N_INSTRUCTION_TYPES; type++ ) {
		doPredsOfNode( C, v, /* pushFlag */FALSE,
					type, nFunctionalUnits( type ) );
	    }
	}

	for( p = C->latencySucc[v]; p != NULL; p = p->next ) {
	    w = p->vertex;
	    if( C->lower[w] < (C->lower[v] + p->latency) ) {
		C->lower[w] =  C->lower[v] + p->latency;
	    }
	}

	for( p = C->distanceSucc[v]; p != NULL; p = p->next ) {
	    w = p->vertex;
	    if( C->lower[w] < (C->lower[v] + p->latency) ) {
		C->lower[w] =  C->lower[v] + p->latency;
	    }
	}
    }
}

/*
 *  Set and improve the upper bounds of the variables by
 *  propagating the latency, distance, and edge constraints.
 */
static void
setUpperBounds( CSP_type *C, int order[], int m )
{
    int  i, v, w, type;
    ELEM *p;

    C->m = m;
    for( v = C->n; v >= 1; v-- ) {
	C->upper[v] = m;
    }

    for( i = C->n; i >= 1; i-- ) {
	v = order[i];

	if( C->nSuccs[v] > 1 ) {
	    for( type = 0; type < N_INSTRUCTION_TYPES; type++ ) {
		doSuccsOfNode( C, v, /* pushFlag */FALSE,
					type, nFunctionalUnits( type ) );
	    }
	}

	for( p = C->latencyPred[v]; p != NULL; p = p->next ) {
	    w = p->vertex;
	    if( C->upper[w] > (C->upper[v] - p->latency) ) {
		C->upper[w] =  C->upper[v] - p->latency;
	    }
	}
	for( p = C->distancePred[v]; p != NULL; p = p->next ) {
	    w = p->vertex;
	    if( C->upper[w] > (C->upper[v] - p->latency) ) {
		C->upper[w] =  C->upper[v] - p->latency;
	    }
	}
    }
}

/*
 *  Increment the upper bounds of the variables,
 *  ensuring that the first and the last variable have
 *  their domains equal to [1,1] and [m,m], respectively.
 */
void
incrementUpperBounds( CSP_type *C )
{
    int  v;

    C->m++;
    for( v = 1; v <= C->n; v++ ) {
	if( !C->fixed[v] ) {
	    C->upper[v]++;
	}
    }

    C->lower[1]    = C->upper[1]    = 1;
    C->lower[C->n] = C->upper[C->n] = C->m;
}

void
addToUpperBounds( CSP_type *C, int offset )
{
    int  v;

    C->m += offset;
    for( v = 1; v <= C->n; v++ ) {
	if( !C->fixed[v] ) {
	    C->upper[v] += offset;
	}
    }

    C->lower[1] = C->upper[1] = 1;
}

/*
 *  Set the initial lower and upper bounds of the variables.
 *  ensuring that the first and the last variable have
 *  their domains equal to [1,1] and [m,m], respectively.
 */
void
setBounds( CSP_type *C )
{
    int  m;
    int  *order;
    int  index;
    BOOL *marked;

    order = (int *)calloc( (size_t)C->n+1, sizeof(int) );
    marked = (BOOL *)calloc( (size_t)C->n+1, sizeof(BOOL) );

    index = C->n;
    topologicalSort( C, 1, order, &index, marked );

    setLowerBounds( C, order );

    m = C->lower[C->n];
    setUpperBounds( C, order, m );

    C->lower[1] = C->upper[1] = 1;
    C->lower[C->n] = C->upper[C->n] = m;

    free( order );
    free( marked );
}

/*
 *  Save the lower and upper bounds of the variables at
 *  currentLevel.
 */
void
saveBounds( CSP_type *C, int currentLevel )
{
    int  v;

    if( currentLevel > C->n ) {
	fprintf( stderr, "saveBounds out of range %d > %d\n",
	    currentLevel, C->n );
    }

    for( v = 1; v <= C->n; v++ ) {
	C->saveLower[currentLevel][v] = C->lower[v];
	C->saveUpper[currentLevel][v] = C->upper[v];
    }
}

/*
 *  Restore the lower and upper bounds of the variables at
 *  currentLevel.
 */
void
restoreBounds( CSP_type *C, int currentLevel )
{
    int  v;

    for( v = 1; v <= C->n; v++ ) {
	C->lower[v] = C->saveLower[currentLevel][v];
	C->upper[v] = C->saveUpper[currentLevel][v];
    }
}

/* =================================================================
 *  Routines for printing/debugging/viewing.
 */

void
printCSP( CSP_type *C )
{
    int  v;
    ELEM *p;

    printf( "\tn: %d\n", C->n );
    printf( "\tm: %d\n\n",  C->m );
    for( v = 1; v <= C->n; v++ ) {
	printf( "\t%3d: [%3d,%3d] = %3d; pred=%3d; succ=%3d; type=%d; d=%d, serial=%d\n",
	    v,
	    C->lower[v],
	    C->upper[v],
	    C->upper[v] - C->lower[v] + 1,
	    C->nPreds[v],
	    C->nSuccs[v],
	    C->varType[v],
#ifdef USE_PIPELINING
		C->varDuration[v],
#else
		1,
#endif

#ifdef USE_EXECUTION_SERIALIZATION
		C->varIsSerial[v]
#else
		0
#endif
		);
    }
    printf( "\n" );

    printf( "Successors:\n" );
    for( v = 1; v <= C->n; v++ ) {
	printf( "%d", v );
	for( p = C->latencySucc[v]; p != NULL; p = p->next ) {
	    printf( " %d:%d", p->vertex, p->latency );
	}
	printf( "\n" );
    }
    printf( "\n" );

    printf( "Predecessors:\n" );
    for( v = 1; v <= C->n; v++ ) {
	printf( "%d", v );
	for( p = C->latencyPred[v]; p != NULL; p = p->next ) {
	    printf( " %d:%d", p->vertex, p->latency );
	}
	printf( "\n" );
    }
    printf( "\n" );
}

void
printSolution( CSP_type *C )
{
    int  t, v;

    for( t = 1; t <= C->m; t++ ) {
	printf( "Instruction(s) at time %d:", t );
	for( v = 1; v <= C->n; v++ ) {
	    if( C->lower[v] == t && C->upper[v] == t ) {
		printf( " %d(%d)", v-1, C->varType[v] );
	    }
	}
	printf( "\n" );
    }
}

/*
 *  Print the CSP graph in a form suitable for viewing by yEd.
 */
void
printGraph( CSP_type *C )
{
    int	 v;
    ELEM *q;

    printf("graph\n");
    printf("[\n");
    printf("\tdirected\t1\n");

    for( v = 1; v <= C->n; v++ ) {
	printf("\tnode\n");
	printf("\t[\n");
	printf("\t\tid\t%d\n", v);
	printf("\t\tlabel\t\"%d\"\n", v);
	printf("\t\tgraphics\n");
	printf("\t\t[\n");
	printf("\t\t\ttype\t\"rectangle\"\n");
	switch( C->varType[v] ) {
	    case 0: printf("\t\t\tfill\t\"#FF0000\"\n");
		    break;
	    case 1: printf("\t\t\tfill\t\"#8EFF00\"\n");
		    break;
	    case 2: printf("\t\t\tfill\t\"#002DFF\"\n");
		    break;
	    case 3: printf("\t\t\tfill\t\"#FFFF00\"\n");
		    break;
	    case 4: printf("\t\t\tfill\t\"#FFFFFF\"\n");
		    break;
	}
	printf("\t\t]\n");
	printf("\t]\n");
    }

    for( v = 1; v <= C->n; v++ ) {
	for( q = C->latencySucc[v]; q != NULL; q = q->next ) {
	    printf("\tedge\n");
	    printf("\t[\n");
	    printf("\t\tsource\t%d\n", v);
	    printf("\t\ttarget\t%d\n", q->vertex);
	    printf("\t\tlabel\t\"%d\"\n", q->latency);
	    printf("\t\tgraphics\n");
	    printf("\t\t[\n");
	    printf("\t\t\tfill\t\"#000000\"\n");
	    printf("\t\t\ttargetArrow\t\"standard\"\n");
	    printf("\t\t]\n");
	    printf("\t]\n");
	}
    }
    printf("]\n");
    fflush(stdout);
}

void
verifySolution( CSP_type *C )
{
    int  t, v, type, status, *count;
    ELEM *p;

#ifdef USE_PIPELINING
	int room;
#endif

#ifdef DEBUG
	printSolution( C );
#endif
    
	count = (int *)calloc( (size_t)C->m+1, sizeof(int) );

    status = 1;
    for( v = 1; v <= C->n; v++ ) {
	status = status && (C->lower[v] == C->upper[v]);
    }
    if( !status ) {
	printf( "Verifying CSP Scheduler solution:\n" );
	printf( "\tEach instruction does not have a unique value.\n" );
	exit( 6 );
    }

    for( type = 0; type < N_INSTRUCTION_TYPES; type++ ) {
	for( t = 1; t <= C->m; t++ ) {
	    count[t] = 0;
	}
	for( v = 1; v <= C->n; v++ ) {
	    if( C->varType[v] == type )
		count[C->lower[v]]++;
	}
	status = 1;
	for( t = 1; t <= C->m; t++ ) {
	    status = status && (count[t] <= nFunctionalUnits( type ));
	}
	if( !status ) {
	    printf( "Verifying CSP Scheduler solution:\n" );
	    printf( "\tIssue-width constraint on type %d not satisfied.\n", type );
	    exit( 6 );
	}
    }

	/* Check overall issue width constraint */
#ifdef USE_ISSUE_WIDTH
	for( t = 1; t <= C->m; t++ ) {
	    count[t] = 0;
	}
	for( v = 1; v <= C->n; v++ ) {
		if( C->varType[v] != PSEUDO_INSTR )
			count[C->lower[v]]++;
	}
	status = 1;
	for( t = 1; t <= C->m; t++ ) {
	    status = status && (count[t] <= getIssueWidth());
	}
	if( !status ) {
	    printf( "Verifying CSP Scheduler solution:\n" );
	    printf( "\tOverall issue-width constraint not satisfied.\n" );
	    exit( 6 );
	}
#endif
    
	status = 1;
    for( v = 1; v <= C->n; v++ ) {
	for( p = C->latencySucc[v]; p != NULL; p = p->next ) {
	    status = status &&
			((C->lower[v] + p->latency) <= C->lower[p->vertex]);
	}
    }
    if( !status ) {
	printf( "Verifying CSP Scheduler solution:\n" );
	printf( "\tSuccessor constraints not satisfied.\n" );
	exit( 6 );
    }

    status = 1;
    for( v = 1; v <= C->n; v++ ) {
	for( p = C->latencyPred[v]; p != NULL; p = p->next ) {
	    status = status &&
			((C->lower[p->vertex] + p->latency) <= C->lower[v]);
	}
    }
    if( !status ) {
	printf( "Verifying CSP Scheduler solution:\n" );
	printf( "\tPredecessor constraints not satisfied.\n" );
	exit( 6 );
    }

#ifdef USE_PIPELINING
	// Step through time, making sure that there's enough resources at each cycle
	for( type = 0; type < N_INSTRUCTION_TYPES; type++ ) {
		for( t = 1; t <= C->m; t++ ) {
			room = nFunctionalUnits( type );
			for( v = 1; v <= C->n; v++ ) {
				if( C->varType[v] == type ) {
					if( C->lower[v] == t ) {
						--room;
					}
					else if( C->lower[v] < t ) {
						if( C->varDuration[v] > 1 && C->lower[v] + C->varDuration[v] - 1 >= t ) {
							--room;
						}
					}

					if( room < 0 ) {
						printf( "Verifying CSP Scheduler solution:\n" );
						printf( "\tPipelining constraint on type %d not satisfied in cycle %d.\n", type, t );
						exit( 6 );
					}
				}
			}
		}
	}

	// Also make sure all instructions finish before the final cycle
	for( v = 1; v <= C->n; v++ ) {
		if( C->lower[v] + C->varDuration[v] - 1 > C->m ) {
			printf( "Verifying CSP Scheduler solution:\n" );
			printf( "\tInstruction %d completes after the final cycle.\n", v );
			exit( 6 );
		}
	}
#endif

#ifdef USE_EXECUTION_SERIALIZATION
	for( v = 1; v <= C->n; v++ ) {
		// If this instruction is execution-style, check it
		if( C->varIsSerial[v] ) {
			for( t = 1; t <= C->n; t++) {
				// Don't compare an instruction against itself; this will cause failure :)
				if (t == v || C->varType[t] == PSEUDO_INSTR) {
					continue;
				}

				// Instruction t can't start before v and end after
				if ( C->lower[v] > C->lower[t] ) {
					if ( C->lower[t] + C->varDuration[t] - 1 > C->lower[v] ) {
						printf( "Verifying CSP Scheduler solution:\n" );
						printf( "\tExecution-style instruction wasn't the last incomplete instruction when it began execution.\n" );
						exit( 6 );
					}
				}
				// Instruction also can't begin in the same cycle as v
				else if ( C->lower[v] == C->lower[t] ) {
					printf( "Verifying CSP Scheduler solution:\n" );
					printf( "\tMore than one instruction issued in a cycle when an execution-style instruction was issued\n" );
					//exit( 6 );
				}
			}
		}
	}
#endif
    
	free( count );
}

/*
 * Specify a timeout for a given timeout level
 */
void
setCSPTimeout( int level, double timeLimit )
{
	assert( level >= 0 && level < NUM_TIMEOUT_LEVELS );
	assert( timeLimit > 0.0 );

	timeoutLevels[level] = timeLimit;
}

/*
 * Get the timeout value for a given level
 */
double
getCSPTimeout( int level )
{
	assert( level >= 0 && level < NUM_TIMEOUT_LEVELS );

	return timeoutLevels[level];
}

/*
 * Set the max region size
 */
void
setMaxRegionSize( int size )
{
	assert( size > 0 );

	maxRegionSize = size;
}

/*
 * Get the max region size
 */
int
getMaxRegionSize()
{
	return maxRegionSize;
}

/*
 * Set the max restart size
 */
void
setMaxRestartSize( int size )
{
	maxRestartSize = size;
}

/*
 * Get the max restart size
 */
int
getMaxRestartSize()
{
	return maxRestartSize;
}
