#ifndef _CSP_H_
#define _CSP_H_

#include "dag.h"
#include "list.h"
#include "stack.h"
#include "processor.h"
#include "propagate.h"

/* =================================================================
 *  Abstract data type to represent a CSP.
 */
typedef struct {
    /*
     *  Variables.
     */
    int  n;		/* number of variables; their names are 1,...,n */
    /*
     *  Domains.
     */
    int  m;		/* domains are in the range [1, ..., m]		*/
    int  *lower;	/* current lower bound on variable v		*/
    int  *upper;	/* current upper bound on variable v		*/
    /*
     *  Constraints.
     *  For a description of a constraint, see the comments
     *  for the corresponding routine that sets that constraint.
     */
    ELEM **latencySucc;
    ELEM **latencyPred;
    ELEM **distanceSucc;
    ELEM **distancePred;
    ELEM **distanceUpperSucc;
    ELEM **distanceUpperPred;
    ELEM **dominanceSucc;
    ELEM **dominancePred;
    ELEM **edgeSucc;
    ELEM **edgePred;
    /*
     *  Cost function.
     */
    int  nSideExits;	/* number of side exits				*/
    int  *costCoeffs;	/* coefficients of the cost function		*/
    double *costFreqs;	/* execution frequencies of the cost function	*/
    int  *costVars;	/* variables of the cost function		*/
    int  lbCost;	/* lower bound on cost of optimal schedule	*/
    int  ubCost;	/* upper bound on cost of optimal schedule	*/
    /*
     *  Properties.
     */
    int  *varType;	/* operation type of each variable/instruction	*/
    BOOL *fixed;	/* is the value of this variable fixed?		*/
    BOOL *isArticulationNode;
			/* is the variable an articulation node?	*/

#ifdef USE_PIPELINING
	int  *varDuration; /* Duration of each instruction */
	int   nDummyNodes; /* Number of dummy nodes */
#endif

#ifdef USE_EXECUTION_SERIALIZATION
	BOOL *varIsSerial; /* True if variable needs serial execution */
#endif

    int  *nOps;		/* number of operations of each type		*/
    int  *nPreds;	/* number of immediate predecessors of v	*/
    int  *nSuccs;	/* number of immediate successors of v		*/
    int  **cp;		/* critical path distances between variables	*/
    BOOL **edgeMarked;  /* used to prevent cycles in dominance cons.    */
    /*
     *  Solving/search.
     */
    /*
     *  Statistics.
     */
    int	 visits;	/* Nodes visited so far during search		*/
    int	 backtracks;	/* Backtracks performed so far during search	*/
    /*
     *  Saving/restoring.
     *  Space for storing (and restoring) the lower and upper bounds
     *  of the variables at the 0, ..., n levels of the search.
     */
    int  **saveLower;
    int  **saveUpper;
    /*
     *  Variable ordering heuristic.
     */
    int  *changev;	/* amount of propagation this variable caused when */
			/*     probing was used to test lower bound        */
    /*
     *  Propagating.
     */
    int		propagationLevel;/* 0 means propagate, > 0 means probing */
    int		changes;	/* domains changed during last propagation */
    BOOL	iterNeeded[N_INSTRUCTION_TYPES];
    STACK_t	*s_lower;
    STACK_t	*s_upper;
    node_type	*values;	/* used by doPred, doSucc */
    BOOL	*filled;	/* used by BinaryGCC	  */
    int		*t;		/* used by gcc ...	  */
    int		*d;
    int		*h;
    interval	*iv;
    int		niv;
    interval	**minsorted;
    interval	**maxsorted;
    int		*bounds;
    int		nb;
    /*
     *  Timing and restarts.
     */
    double	startTime;
    double	maxCPUseconds;
    BOOL	restartsAllowed;
} CSP_type;

/*
 *  Possible values returned by solve and probe routines.
 */
#define	  NO_SOLUTION	  0	/* no solution exists		 	 */
#define	  PROVED_OPTIMAL  1	/* heuristic solution was proven optimal */
#define	  SOLUTION	  2	/* a solution was found			 */
#define	  TIMED_OUT	  3	/* unknown as resource limit exceeded 	 */
#define	  OUT_OF_SPACE	  4	/* not enough space for enumerateAndTest */

/*
 *  When enumerating the assignments to the cost variables
 *  which satisfy given bounds on the cost function, maximum
 *  number of assignments to store.
 */
#define MAX_NUMBER_ASSIGNS  1024
//#define MAX_NUMBER_ASSIGNS  102400
//#define MAX_NUMBER_ASSIGNS  409600

/* =================================================================
 *  Routines for creating and destroying a CSP with n variables.
 */
CSP_type *createCSP( int n );
void	 destroyCSP( CSP_type *C );

/* =================================================================
 *  Routines for manipulating the bounds of the domains of the
 *  variables.
 */
void	setBounds( CSP_type *C );
void	incrementUpperBounds( CSP_type *C );
void	addToUpperBounds( CSP_type *C, int offset );
void	saveBounds( CSP_type *C, int currentLevel );
void	restoreBounds( CSP_type *C, int currentLevel );

/* =================================================================
 *  Routines for the cost function.
 */
void	setCostFunction( CSP_type *C, DAG_type *DAG );
void	printCostFunction( CSP_type *C );
int	getCost( CSP_type *C );
int	getCostSchedule( CSP_type *C, int schedule[] );
int	getLengthLowerBound( CSP_type *C, int j );
int	getLengthUpperBound( CSP_type *C, int j );
void    enumerateCostVariables( CSP_type *C, int **all, int *n_assignments,
		int *partial, int current );
void	setArticulationNodes( CSP_type *C );
int	allArticulationNodes( CSP_type *C );

/* =================================================================
 *  Routine for setting the constraints over the variables.
 */
void	setConstraints( CSP_type *C, DAG_type *DAG );

/* =================================================================
 *  Routines for propagating constraints.
 */
BOOL	doPredsOfNode( CSP_type *C, int v, BOOL pushFlag, int type, int width );
BOOL	doSuccsOfNode( CSP_type *C, int v, BOOL pushFlag, int type, int width );
BOOL	propagate( CSP_type *C, int currentVar, int currentLevel );
void    safePrune( CSP_type *C );

/* =================================================================
 *  Routines for printing/debugging/viewing.
 */
void	printCSP( CSP_type *C );
void	printSolution( CSP_type *C );
void	printGraph( CSP_type *C );
void	verifySolution( CSP_type *C );

/*
 *  Timeouts for two stages of solving (without probing, with probing).
 */
#define   NUM_TIMEOUT_LEVELS 3
#define   TIME_LIMIT_FIRST	 0
#define   TIME_LIMIT_SECOND  1
#define   TIME_LIMIT_TOTAL   2

void   setCSPTimeout( int level, double timeLimit );
double getCSPTimeout( int level );

/* Max region and restart sizes */
void setMaxRegionSize( int size );
int  getMaxRegionSize();
void setMaxRestartSize( int size );
int  getMaxRestartSize();

/* =================================================================
 *  Routines for timing and restarts.
 */
void	set_start_time( CSP_type *C );
void	set_absolute_CPU_limit( CSP_type *C, double time_limit );
void	set_relative_CPU_limit( CSP_type *C, double time_limit );
BOOL	limit_exceeded( CSP_type *C );


#endif
