
#include <string.h>

#include "global.h"
#include "processor.h"
#include "list.h"
#include "dag.h"
#include "instr.h"
#include "settings.h"


static BOOL isBranchPoint( char instructionName[] );
static BOOL isBreakableEdge( char edgeType[] );


/*
 *  Read in and store the DAG.
 */
DAG_type *
getDAG( char *problem, BOOL doBreakEdges )
{
    DAG_type    *DAG;
    int     i, j, v_i, v_j, latency, n, k, percent;
    int     *nPred, *succs;
    char    name[MAX_LINE_LENGTH + 1 + 256 + 1];
    char        line[MAX_LINE_LENGTH + 1];
    FILE    *fp;
    char    temp[MAX_LINE_LENGTH + 1];
    char    instructionName[MAX_LINE_LENGTH + 1], edgeType[MAX_LINE_LENGTH + 1];
    double      a, b;

    sprintf( name, "%s%s", getBaseDataPath(), problem );
    if( (fp = fopen( name, "r" )) == NULL ) {
    fprintf( stderr, "Could not open file %s\n", name );
    exit( 6 );
    }

    DAG = (DAG_type *)calloc( (size_t)1, sizeof(DAG_type) );

    fgets( line, MAX_LINE_LENGTH, fp );   // skip: ID:benchmark:file:block
    fscanf( fp, "NN:%d\n", &n);       // read: NN:nodes

    DAG->nInstructions = n+1;         // add one for dummy root at 0 (node for the dummy sink added later)
    k = DAG->nInstructions;
    DAG->blockID = problem;
                          // +1 as may need to add a dummy sink node.
    DAG->instructionType = (int *)calloc( (size_t)DAG->nInstructions+1, sizeof(int) );
    DAG->percentage = (int *)calloc( (size_t)DAG->nInstructions+1, sizeof(int) );
    DAG->f1 = (double *)calloc( (size_t)DAG->nInstructions+1, sizeof(double) );
    DAG->f2 = (double *)calloc( (size_t)DAG->nInstructions+1, sizeof(double) );
    DAG->successors = (ELEM **)calloc( (size_t)DAG->nInstructions+1, sizeof(ELEM *) );
    nPred = (int *)calloc( (size_t)DAG->nInstructions+1, sizeof(int) );
    for( i = 0; i <= DAG->nInstructions; i++ ) {
    DAG->successors[i] = (ELEM *)NULL;
    nPred[i] = 0;
    }
    succs = (int *)calloc( (size_t)DAG->nInstructions+1, sizeof(int) );
    DAG->instructionIndex = (int *)calloc( (size_t)DAG->nInstructions + 1, sizeof(int) );
    DAG->nDummyNodes = 0;

    fgets( line, MAX_LINE_LENGTH, fp ); /* read: Node ... */

START:
    /*
     *  Node ...
     */
    i = 0;
    line[4] = ' '; // Ensure that we have Node N, for some int N
    while( line[i] != ' ' ) {
        i++;
    }
    v_i = atoi( &line[i+1] );
    if( (v_i <= 0) || (v_i >= DAG->nInstructions) ) {
        fprintf( stderr, "v_i out of range: %d > %d.\n",
        v_i, DAG->nInstructions );
        exit( 6 );
    }
    for( i = 1; i <= DAG->nInstructions; i++ ) {
        succs[i] = -1;
    }

    /*
     *  Assembler instruction ...
     */
    fgets( line, MAX_LINE_LENGTH, fp );
    strcpy(temp, line);

    // Get the registers and the instruction time from the line
    i = 0;
    while( line[i] != ' ' ) {
        i++;
    }
    line[i] = 0;
    DAG->instructionIndex[v_i] = getIndex(line);
    if( instructionTimes[DAG->instructionIndex[v_i]][0] > 1 ) {
        DAG->nDummyNodes += instructionTimes[DAG->instructionIndex[v_i]][0] - 1;
    }
    line[i] = ' ';

    // Get the branch taken, if a branch instruction
    strcpy(line, temp);
    for( i = 0; line[i] != ' '; i++ ) {
        instructionName[i] = line[i];
    }
    instructionName[i] = '\0';
    if( isBranchPoint( instructionName ) ) {
        j = strlen( line ) - 1;
        while( (j > 0) &&
           (line[j] != '*') &&
           (line[j] != '=') ) {
        j--;
        }
        if( j == 0 ) {
            fprintf( stderr, "cannot find taken percent\n" );
            exit( 6 );
        }
        if( line[j] == '*' ) {
        percent = 1;
        a = 0.0;
        b = 0.0;
        }
        else {
        j++;
        percent = atoi(&line[j]);
        while( line[j] != '(' ) {
            j++;
        }
        j++;
        a = atof(&line[j]);
        while( line[j] != ',' ) {
            j++;
        }
        j++;
        b = atof(&line[j]);
        }
        if( percent == 0 ) percent = 1;
        if( (percent <= 0) || (percent > 100) ) {
            fprintf( stderr, "faulty percentage found: %d\n", percent );
            exit( 6 );
        }
        DAG->percentage[v_i] = percent;
        DAG->f1[v_i] = a;
        DAG->f2[v_i] = b;
#ifdef DEBUG
        printf( "Instruction %d: Branch instruction with %d%% chance\n", v_i, percent );
#endif
    }
    else
    {
        a = 0.0;
        b = 0.0;
        DAG->percentage[v_i] = 0;
        DAG->f1[v_i] = 0.0;
        DAG->f2[v_i] = 0.0;
    }

    /*
     *  Instruction type ...
     */
    fgets( line, MAX_LINE_LENGTH, fp ); // Ignore what's in the DAG and use the accurate mapping
    DAG->instructionType[v_i] = getInstructionType( instructionMap[DAG->instructionIndex[v_i]][2] );

    // Handle the case where an instruction isn't found in the map: use the unit from the file
    if( DAG->instructionType[v_i] == -1 ) {
        i = 0;
        while( line[i] != '\n' ) {
            i++;
        }
        line[i] = 0;
        DAG->instructionType[v_i] = getInstructionType( line );
    }

NEXT_EDGE:
    fgets( line, MAX_LINE_LENGTH, fp ); // END or Node or s:v:l
    if( line[0] == 'E' &&
        line[1] == 'N' &&
        line[2] == 'D' ) {
        for( i = 1; i <= DAG->nInstructions; i++ ) if( succs[i] >= 0 ) {
        insertList( &(DAG->successors[v_i]), i, succs[i] );
        }
        goto END;
    }
    if( line[0] == 'N' &&
        line[1] == 'o' &&
        line[2] == 'd' &&
        line[3] == 'e' ) {
        for( i = 1; i <= DAG->nInstructions; i++ ) if( succs[i] >= 0 ) {
        insertList( &(DAG->successors[v_i]), i, succs[i] );
        }
        goto START;
    }

    /*
     *  Process  type : successor : latency
     */
    for( i = 0; line[i] != ':'; i++ ) {
        edgeType[i] = line[i];
    }
    edgeType[i] = '\0';
    i++;
    v_j = atoi( &line[i] );
    if( (v_j < 1) || (v_j >= DAG->nInstructions) ) {
        fprintf( stderr, "v_j out of range: %d > %d.\n",
        v_j, DAG->nInstructions );
        exit( 6 );
    }
    if( v_i >= v_j ) {
        fprintf( stderr, "Violates assumption that all successors\n" );
        fprintf( stderr, "of a vertex have a higher vertex number.\n" );
        fprintf( stderr, "%d >= %d\n", v_i, v_j );
        exit( 6 );
    }
    while( line[i] != ':' ) {
        i++;
    }
    i++;
    latency = atoi( &line[i] );
    if( (latency < 0) || (latency > getMaxLatency()) ) {
        fprintf( stderr, "latency out of range: %d\n", latency );
        exit( 6 );
    }

    if( doBreakEdges && isBreakableEdge( edgeType ) ) {
        // breakable edge
    }
    else {
        nPred[v_j]++;
        if( succs[v_j] < latency ) {
        succs[v_j] = latency;
        }
    }

    goto NEXT_EDGE;

END:
    /*
     *  Ensure there is a single source node by adding
     *  a dummy node 0 and adding an edge with latency
     *  1 between node 0 and any node without predecessors.
     */
    DAG->instructionType[0] = PSEUDO_INSTR;
    for( i = 1; i < DAG->nInstructions; i++ ) {
        if( nPred[i] == 0 ) { // has no predecessors
            insertList( &(DAG->successors[0]), i, 1 );
        }
    }

    /*
     * Add a sink node, with latencies between previous sinks
     * and the new node as the actual latencies for the
     * instructions of the sinks
     */
    DAG->instructionType[DAG->nInstructions] = PSEUDO_INSTR;
    for( i = 1; i < DAG->nInstructions; i++ ) {
        if( DAG->successors[i] == NULL ) { // has no successors
            insertList( &(DAG->successors[i]),
                        DAG->nInstructions,
                        instructionTimes[DAG->instructionIndex[i]][1] );
        }
    }
    DAG->nInstructions++;

#ifdef DEBUG
    printDAG(DAG);
#endif
    
    fclose( fp );
    free( nPred );
    free( succs );

    return( DAG );
}

/*
 *  Print the DAG for debugging purposes.
 */
void
printDAG( DAG_type *DAG )
{
    int  i;
    ELEM *q;

    printf( "\nvertices: %d\n", DAG->nInstructions );

    printf( "Successors:\n" );
    for( i = 0; i < DAG->nInstructions; i++ ) {
    printf( "%d (type: %d, duration: %d", i, DAG->instructionType[i],
        instructionTimes[DAG->instructionIndex[i]][0] );
    if( isExecution( DAG->instructionIndex[i] ) ) {
        printf( ", serial" );
    }

    // Successors
    for( q = DAG->successors[i]; q != NULL; q = q->next ) {
        printf( " %d:%d", q->vertex, q->latency );
    }
    printf( "\n" );
    }
    printf( "\n" );
}


#define MAX_BRANCH_TYPES 8

char *branchMap[MAX_BRANCH_TYPES] = {
    "BCF",
    "BCT",
    "BF",
    "BT",
    "BFBIT",
    "BTBIT",
    "BCTF",
    "BCTT"
};

static BOOL
isBranchPoint( char instructionName[] )
{
    int i;

    for( i = 0; i < MAX_BRANCH_TYPES; i++ ) {
        if( strcmp( instructionName, branchMap[i] ) == 0 ) {
            return( TRUE );
        }
    }
    return( FALSE );
}

/*
#define MAX_BREAKABLE_EDGE_TYPES 3

char *edgeMap[MAX_BREAKABLE_EDGE_TYPES] = {
    "BSPEC",
    "TBUU",
    "TBLRU",
};
*/

#define MAX_BREAKABLE_EDGE_TYPES 8

char *edgeMap[MAX_BREAKABLE_EDGE_TYPES] = {
    "BSPEC",
    "TBUU",
    "TBLRU",
    "AW",
    "TM",
    "TW",
    "W",
    "WM"
};

static BOOL
isBreakableEdge( char edgeType[] )
{
    int i;

    for( i = 0; i < MAX_BREAKABLE_EDGE_TYPES; i++ ) {
        if( strcmp( edgeType, edgeMap[i] ) == 0 ) {
            return( TRUE );
        }
    }
    return( FALSE );
}

