/********************************************************************
* heuristic.c                                                       *
*********************************************************************
* Written by Mike Chase                                             *
* Written on February 18, 2006                                      *
*********************************************************************
* Purpose: Allow various heuristics to be used in the schedulers    *
********************************************************************/

#include <assert.h>
#include "global.h"
#include "list.h"
#include "dag.h"
#include "ls.h"
#include "processor.h"
#include "heuristic.h"


int lsHeuristic = -1;
int optHeuristic = -1;

// Set the heuristic used for the list scheduler
void setLSHeuristic( int heuristic )
{
    assert( heuristic >= 0 && heuristic < HEURISTIC_NUM_HEURISTICS 
        && heuristic != HEURISTIC_BEST );
    lsHeuristic = heuristic;
}

// Set the heuristic used when estimating an upper bound on schedule
// length for the optimal scheduler
void setOptHeuristic( int heuristic )
{
    assert( heuristic >= 0 && heuristic < HEURISTIC_NUM_HEURISTICS );
    optHeuristic = heuristic;
}

int getLSHeuristic( ) { return lsHeuristic; }
int getOptHeuristic( ) { return optHeuristic; }



/*
 *  Find instruction with max delay.
 *  Break ties by lowest EST.
 */
int
findBestInstruction( LS_type *L, int* room, int currentTime, int heuristic )
{
    int v, i;
    int bestInstruction;

    bestInstruction = 0;
    for( i = 0; i < N_INSTRUCTION_TYPES; i++ ) {
        for( v = 1; v <= L->n; v++ ) {
            /*
             *  An instruction v is ready to be scheduled if all
             *  of its predecessors are scheduled (nPreds[v] == 0)
             *  and enough time has passed to satisfy the latency
             *  constraints (EST[v] <= currentTime)
             */
            if( (room[L->varType[v]] > 0 &&
                L->nPreds[v] == 0) &&
                !L->isScheduled[v] &&
                L->EST[v] <= currentTime &&
                ( L->varType[v] == getInstructionTypePreference( i ) ||
                getInstructionTypePreference( i ) == INSTR_DONT_CARE ) )
            {
                if( bestInstruction == 0 )
                {
                    bestInstruction = v;
                }
                else if( better( L, v, bestInstruction, currentTime, heuristic ) )
                {
                    bestInstruction = v;
                }
            }
        }

        /* 
         * Don't do another iteration if we just had a DONTCARE iteration,
         * as this will ensure the best instruction, if any, has been found
         */
        if ( getInstructionTypePreference( i ) == INSTR_DONT_CARE ) {
            break;
        }
    }

    return( bestInstruction );
}

/*
 *  Is v1 a better choice than v2?
 */
int
better( LS_type *L, int v1, int v2, int currentTime, int heuristic )
{
    int max1, max2;
    int latency_sum1, latency_sum2;
    int pl_source1, pl_source2;
    int pl_sink1, pl_sink2;
    int cp_source1, cp_source2;
    int cp_sink1, cp_sink2;
    int slack1, slack2;
    int est1, est2;
    int name1, name2;
    int descendants1, descendants2;
    int succs1, succs2;
    int exec_time1, exec_time2;

    name1 = v1;
    name2 = v2;

    max1 = max( L->distance[v1], L->cp[v1][L->n] );
    max2 = max( L->distance[v2], L->cp[v2][L->n] );

    pl_source1 = L->pl[1][v1];
    pl_source2 = L->pl[1][v2];

    pl_sink1 = L->pl[v1][L->n];
    pl_sink2 = L->pl[v2][L->n];

    cp_source1 = L->cp[1][v1];
    cp_source2 = L->cp[1][v2];

    cp_sink1 = L->cp[v1][L->n];
    cp_sink2 = L->cp[v2][L->n];

    slack1 = L->cp[1][L->n] - (L->cp[1][v1] + L->cp[v1][L->n]);
    slack2 = L->cp[1][L->n] - (L->cp[1][v2] + L->cp[v2][L->n]);

    est1 = L->EST[v1];
    est2 = L->EST[v2];

    latency_sum1 = L->sumLatencies[v1];
    latency_sum2 = L->sumLatencies[v2];

    succs1 = L->nSuccs[v1];
    succs2 = L->nSuccs[v2];

    descendants1 = L->nDescs[v1];
    descendants2 = L->nDescs[v2];

    exec_time1 = L->eTime[v1];
    exec_time2 = L->eTime[v2];


    switch( heuristic )
    {
        case HEURISTIC_CRITICAL_PATH:
        {
            if( L->maxDelay[v1] > L->maxDelay[v2] ) {
                return( 1 );
            }
            else if( (L->maxDelay[v1] == L->maxDelay[v2]) &&
                (L->EST[v1] < L->EST[v2]) ) {
                return( 1 );
            }
            else {
                return( 0 );
            }
        }
        case HEURISTIC_DECISIONTREE:
        {
            if (max1 > max2) { return 1; }
            if (max1 < max2) { return 0; }
            if (max1 == max2)
            {
                if (descendants1 == descendants2)
                {
                    if (latency_sum1 == latency_sum2) { return 0; }
                    if (latency_sum1 >  latency_sum2) { return 1; }
                    if (latency_sum1 <  latency_sum2) { return 0; }
                }
                if (descendants1 > descendants2)
                {
                    if (exec_time1 == exec_time2) { return 1; }
                    if (exec_time1 >  exec_time2) { return 1; }
                    if (exec_time1 <  exec_time2) { return 0; }
                }
                if (descendants1 < descendants2)
                {
                    if (exec_time1 == exec_time2) { return 0; }
                    if (exec_time1 >  exec_time2) { return 1; }
                    if (exec_time1 <  exec_time2) { return 0; }
                }
            }
        }
        case HEURISTIC_SHIEH_PAPACHRISTOU:
        {
            if( cp_sink1 > cp_sink2 ) {
                return( 1 );
            }
            else if( (cp_sink1 == cp_sink2) &&
                (exec_time1 > exec_time2) ) {
                return( 1 );
            }
            else if( (cp_sink1 == cp_sink2) &&
                (exec_time1 == exec_time2) &&
                (v1 < v2) ) {
                return( 1 );
            }
            else {
                return( 0 );
            }
        }
        case HEURISTIC_SHIEH_PAPACHRISTOU_2:
        {
            // Shieh and Papachristou with more tie-breaking
            // This is the version that was used in the paper
            if( cp_sink1 > cp_sink2 ) {
                return( 1 );
            }
            else if( (cp_sink1 == cp_sink2) &&
                (exec_time1 > exec_time2) ) {
                return( 1 );
            }
            else if( (cp_sink1 == cp_sink2) &&
                (exec_time1 == exec_time2) &&
                (succs1 > succs2) ) {
                return( 1 );
            }
            else if( (cp_sink1 == cp_sink2) &&
                (exec_time1 == exec_time2) &&
                (succs1 == succs2) &&
                (v1 < v2) ) {
                return( 1 );
            }
            else {
                return( 0 );
            }
        }
        case HEURISTIC_SHIEH_PAPACHRISTOU_3:
        {
            // Shieh and Papachristou with even more tie-breaking
            // This version didn't lead to improvements
            if( cp_sink1 > cp_sink2 ) {
                return( 1 );
            }
            else if( (cp_sink1 == cp_sink2) &&
                (exec_time1 > exec_time2) ) {
                return( 1 );
            }
            else if( (cp_sink1 == cp_sink2) &&
                (exec_time1 == exec_time2) &&
                (succs1 > succs2) ) {
                return( 1 );
            }
            else if( (cp_sink1 == cp_sink2) &&
                (exec_time1 == exec_time2) &&
                (succs1 == succs2) &&
                (pl_source1 < pl_source2) ) {
                return( 1 );
            }
            else if( (cp_sink1 == cp_sink2) &&
                (exec_time1 == exec_time2) &&
                (succs1 == succs2) &&
                (pl_source1 == pl_source2) &&
                (v1 < v2) ) {
                return( 1 );
            }
            else {
                return( 0 );
            }

        }
        case HEURISTIC_EST:
        {
            // earliest start time heuristic
            if( est1 < est2 ) {
                return( 1 );
            }
            else if( (est1 == est2) &&
                (cp_sink1 > cp_sink2) ) {
                return( 1 );
            }
            else if( (est1 == est2) &&
                (cp_sink1 == cp_sink2) &&
                (v1 < v2) ) {
                return( 1 );
            }
            else {
                return( 0 );
            }
        }
        case HEURISTIC_SIDE_EXIT_FIRST:
        {
            if( (L->isSideExit[v1]) && !(L->isSideExit[v2]) ) {
                return( 1 );
            }
            else
            if( !(L->isSideExit[v1]) && (L->isSideExit[v2]) ) {
                return( 0 );
            }
            else
            if( L->maxDelay[v1] > L->maxDelay[v2] ) {
                return( 1 );
            }
            else
            if( (L->maxDelay[v1] == L->maxDelay[v2]) && (L->EST[v1] < L->EST[v2]) ) {
                return( 1 );
            }
            else {
                return( 0 );
            }
        }
        case HEURISTIC_DHASY1:
        case HEURISTIC_DHASY2:
        case HEURISTIC_DHASY3:
        {
            if(L->dhasy[v1] > L->dhasy[v2])
                return( 1 );
            else
                return( 0 );
        }
        case HEURISTIC_WEIGHTED_ESTIMATE:
        {
            if(L->weight_est[v1] > L->weight_est[v2])
                return( 1 );
            else
                return( 0 );
        }
    }

    return 0;
}
