#include <math.h>
#include "global.h"
#include "processor.h"
#include "dag.h"
#include "ls.h"
#include "heuristic.h"
#include "instr.h"
#include "list.h"
#include "pendingList.h"

/* =================================================================
 *  Routines for creating and destroying an LS with n variables.
 */
static LS_type  *createLS( int n );
static void destroyLS( LS_type *L );
BOOL areInstructionsExecuting(LS_type *L, DAG_type *DAG, int order[], int cycle);
BOOL willDelaySerialNode(int bestInstruction, LS_type *L, DAG_type *DAG, int order[], int cycle);


/* =================================================================
 *  Routines for finding the best instruction to schedule.
 */
static int  estimateRegion( LS_type *L, int s, int t, BOOL *inRegion );
static void setLatencyConstraints( LS_type *L, DAG_type *DAG );
static void setEST( LS_type *L );
static void updateEST( LS_type *L, int v, int issueTime );
static void updatePreds( LS_type *L, int v );
static void setMaxDelay( LS_type *L );
static void     setCostCoeffs( LS_type *L, DAG_type *DAG );
static void     printCostCoeffs( LS_type *L );
static void setSideExits( LS_type *L );
static void     calcDhasy1( LS_type *L );
static void     calcDhasy2( LS_type *L );
static void     calcDhasy3( LS_type *L );
static void     calcWeightedEstimate( LS_type *L );
static void criticalPathDistances( LS_type *L, DAG_type *DAG );
static void setEstimatedDistances( LS_type *L, DAG_type *DAG );
static void setPLdistances( LS_type *L );
static void setCPdistances( LS_type *L );
static void setETime( LS_type *L );
static void setSumLatencies( LS_type *L );
static void setDescendants( LS_type *L );
static void setDistances( LS_type * L );

/* =================================================================
 *  Routines for printing/debugging.
 */
static void printLS( LS_type *L );
static void verifySolution( LS_type *L, int order[], DAG_type *DAG );
static int getLSCost( int order[], LS_type *L );
static double getFreqCost( int order[], LS_type *L );

// the following routine was added on August 30, 2007
// in an effort to find a bug where the optimal scheduler
// gave an improved scheduled over that of the list scheduler
// when it shouldn't
static int
findPseudoInstruction( LS_type *L, int currentTime )
{
int v;

    for( v = 1; v <= L->n; v++ ) {
        if( L->varType[v] == PSEUDO_INSTR &&
            L->nPreds[v] == 0 &&
            !L->isScheduled[v] &&
            L->EST[v] <= currentTime ) {
            return( v );
        }
    }

    return( 0 );
}

void listScheduler( DAG_type *DAG, int heuristic, int order[], int *finalCost, int *finalLength, double *costWithFreq )
{
    int      v, type;
    int      bestInstruction = 0, room[N_INSTRUCTION_TYPES];
    int      instructionsToSchedule, currentTime;
    int      instructionsIssued;
    BOOL     anotherPassNeeded;
    LS_type  *L;
#ifdef USE_PIPELINING
    // The cycle in which each instruction is no longer locking the functional unit
    int *lastExecutionCycles, i;
    int duration, pipelineFilledSlots[N_INSTRUCTION_TYPES];
    LISTNODE* node;

    // Initialize lastExecutionCycles
    lastExecutionCycles = (int *)calloc( (size_t)DAG->nInstructions + 1, sizeof(int) );
    for ( i = 1; i <= DAG->nInstructions; i++ ) {
        lastExecutionCycles[i] = -1;
    }

        createPendingList();
#endif

    L = createLS( DAG->nInstructions );
    for( v = 1; v <= L->n; v++ ) {
        order[v-1] = -1;
        L->varType[v] = DAG->instructionType[v-1];
        L->isScheduled[v] = FALSE;
    }

    L->blockID = DAG->blockID;
    setCostCoeffs( L, DAG );
    //printCostCoeffs(L);
    setLatencyConstraints( L, DAG );
    setEST( L );
    setMaxDelay( L );
    setSideExits( L );
    if( heuristic == HEURISTIC_DHASY1 ) {
    setDistances(L);
    calcDhasy1(L);
    }
    if( heuristic == HEURISTIC_DHASY2 ) {
    criticalPathDistances( L, DAG );
    calcDhasy2( L );
    }
    if( heuristic == HEURISTIC_DHASY3 ) {
    criticalPathDistances( L, DAG );
    calcDhasy3( L );
    }
    if( heuristic == HEURISTIC_WEIGHTED_ESTIMATE ) {
    criticalPathDistances( L, DAG );
    setEstimatedDistances( L, DAG );
    calcWeightedEstimate( L );
    }
    setPLdistances( L );
    setCPdistances( L );
    setETime( L );
    setSumLatencies( L );
    setDescendants( L );
    setDistances( L );
    //printLS( L );

    currentTime = 0;
    instructionsToSchedule = L->n;
    while( instructionsToSchedule > 0 ) {
        currentTime++;

        // Set the maximum instructions allowed for each functional unit
        for( type = 0; type < N_INSTRUCTION_TYPES; type++ ) {
            room[type]=nFunctionalUnits( type );
        }

#ifdef USE_PIPELINING
        // First, zero the pipelineFilledSlots array
        for( type = 0; type < N_INSTRUCTION_TYPES; type++ ) {
            pipelineFilledSlots[type] = 0;
        }

        // If the unit is locked, can't schedule
        for( i = 1; i <= DAG->nInstructions; i++ ) {
            if ( lastExecutionCycles[i] >= currentTime ) {
                pipelineFilledSlots[L->varType[i]]++;
            }
        }
#endif

        // Set the total number of instructions that can be issued this cycle
        instructionsIssued = getIssueWidth();

        // Before we schedule new instructions, see if there's anything
        // to mark as "unscheduled"---any instruction that was previously
        // selected but couldn't be issued because of a resource conflict.
        // Empties pendingList, which means the loop after this one will
        // have no effect.
#ifdef USE_RESOURCE_MODEL
        while ( !isPendingListEmpty() ) {
            node = getPendingListHead();
            while( node != NULL ) {
                                L->isScheduled[node->instruction] = FALSE;
                                removeFromPendingList( node->instruction );
                                node = getPendingListHead();
            }
        }
#endif

        // Before we find new instructions, see if there's anything to schedule
        // that was previously selected but couldn't be issued
#ifdef USE_PIPELINING
        anotherPassNeeded = TRUE;
        while ( !isPendingListEmpty() && anotherPassNeeded && instructionsIssued > 0 ) {
            anotherPassNeeded = FALSE;

            // Try to schedule as much as possible from the pending list
            node = getPendingListHead();

            // Walk through the list, trying to schedule the first available node
            while( node != NULL ) {
#ifdef USE_EXECUTION_SERIALIZATION
                if( isExecution( DAG->instructionIndex[node->instruction - 1] ) && 
                    areInstructionsExecuting(L, DAG, order, currentTime) ) {
                    node = node->next;
                    continue;
                }
                else if( !isExecution( DAG->instructionIndex[node->instruction - 1] ) &&
                    willDelaySerialNode( node->instruction, L, DAG, order, currentTime ) ) {
                    node = node->next;
                    continue;
                }
#endif

                // Either a serial node that can be scheduled now, or a non-serial node that
                // doesn't delay a serial node
                // Try to schedule this node: only need resources to be available
                if( room[L->varType[node->instruction]] -
                    pipelineFilledSlots[L->varType[node->instruction]] > 0 ) {
                    // Space available; schedule the instruction
                    bestInstruction = node->instruction;
                    --instructionsIssued;

                    order[bestInstruction-1] = currentTime;
                    room[L->varType[bestInstruction]]--;
                    L->isScheduled[bestInstruction] = TRUE;
                    instructionsToSchedule--;
                    //L->nOps_d[type]--;
                    updatePreds( L, bestInstruction );
                    updateEST( L, bestInstruction, currentTime );

                    // Tie up the functional unit for the instruction
                    duration = instructionTimes[DAG->instructionIndex[bestInstruction - 1]][0];
                    lastExecutionCycles[bestInstruction] = currentTime + duration - 1;

                    // Don't count pseudo-isntructions towards total
                    // (i.e. ignore global issue width constraint)
                    if( L->varType[bestInstruction] == PSEUDO_INSTR )
                        ++instructionsIssued;

                    // Remove the item from the pending list and start over
                    removeFromPendingList( bestInstruction );
                    anotherPassNeeded = TRUE;

#ifdef USE_EXECUTION_SERIALIZATION
                    if( isExecution( DAG->instructionIndex[bestInstruction - 1] ) ) {
                        // We just scheduled a serial node; can't schedule anything else
                        // this cycle
                        anotherPassNeeded = FALSE;
                        instructionsIssued = 0;
                        break;
                    }
#endif
                    break;
                }
                else { // Can't schedule
                    node = node->next;
                }
            }
        }
#endif


        //
        //  Make scheduling passes until no instructions are scheduled
        //
        do {
            anotherPassNeeded = FALSE;
            for( ; instructionsIssued > 0; instructionsIssued--  ) {
                bestInstruction = findBestInstruction( L, room, currentTime, heuristic );
                if( bestInstruction == 0 ) break;

                // Check serial constraints
#ifdef USE_EXECUTION_SERIALIZATION
                if( isExecution( DAG->instructionIndex[bestInstruction - 1] ) && 
                    areInstructionsExecuting(L, DAG, order, currentTime) ) {
                    // Instruction is serial, but something else is already running
                    // this cycle

                    // Put the instruction on the list
                    addToBackOfPendingList( bestInstruction, INSTRUCTION_TYPE_SERIAL );

                    // Flag the instruction as scheduled so that findBestInstruction won't
                    // repeatedly try to select it
                    L->isScheduled[bestInstruction] = TRUE;

                    // We didn't actually issue anything here, so don't actually change
                    // remaining issue slots
                    ++instructionsIssued; // Add one to counter the subtract one in the for loop condition

                    continue;
                }
                else if( !isExecution( DAG->instructionIndex[bestInstruction - 1] ) &&
                    willDelaySerialNode( bestInstruction, L, DAG, order, currentTime ) ) {
                    // Instruction isn't serial, but if we issue it now, we delay a serial instruction
                    // in the pending list

                    // Put this instruction on the list
                    addToBackOfPendingList( bestInstruction, INSTRUCTION_TYPE_DELAYS_SERIAL );

                    // Flag the instruction as scheduled so that findBestInstruction won't
                    // repeatedly try to select it
                    L->isScheduled[bestInstruction] = TRUE;

                    // We didn't actually issue anything here, so don't actually change
                    // remaining issue slots
                    ++instructionsIssued; // Add one to counter the subtract one in the for loop condition

                    continue;
                }
#endif

                // Check pipelining constraints
#ifdef USE_PIPELINING
                if( room[L->varType[bestInstruction]] - pipelineFilledSlots[L->varType[bestInstruction]] == 0 ) {
                    // Put the instruction on the list due to pipeline constraints
                    addToBackOfPendingList( bestInstruction, INSTRUCTION_TYPE_PIPELINED );

                    // Flag the instruction as scheduled so that findBestInstruction won't
                    // repeatedly try to select it
                    L->isScheduled[bestInstruction] = TRUE;

                    // We didn't actually issue anything here, so don't actually change
                    // remaining issue slots
                    ++instructionsIssued; // Add one to counter the subtract one in the for loop condition

                    continue;
                }
#endif

                // All is fine; the node can be scheduled
                anotherPassNeeded = TRUE;

                order[bestInstruction-1] = currentTime;
                room[L->varType[bestInstruction]]--;
                L->isScheduled[bestInstruction] = TRUE;
                instructionsToSchedule--;
                //L->nOps_d[type]--;
                updatePreds( L, bestInstruction );
                updateEST( L, bestInstruction, currentTime );
#ifdef USE_PIPELINING
                // Tie up the functional unit for the instruction
                duration = instructionTimes[DAG->instructionIndex[bestInstruction - 1]][0];
                lastExecutionCycles[bestInstruction] = currentTime + duration - 1;
#endif

                // Don't count pseudo-isntructions towards total
                // (i.e. ignore global issue width constraint)
                if( L->varType[bestInstruction] == PSEUDO_INSTR ) {
                    ++instructionsIssued;
                }

#ifdef USE_EXECUTION_SERIALIZATION
                if( isExecution( DAG->instructionIndex[bestInstruction - 1] ) ) {
                    // Can't issue anything else this cycle: force the loops to
                    // terminate now instead of processing anything else
                    instructionsIssued = 0;
                    anotherPassNeeded = FALSE;
                }
#endif
            }
        } while( anotherPassNeeded );

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
		    // Schedule as many pseudo instructions as possible
		    while( (v = findPseudoInstruction( L, currentTime )) != 0 ) {
                //anotherPassNeeded = TRUE;
                order[v-1] = currentTime;
                L->isScheduled[v] = TRUE;
                instructionsToSchedule--;
                updatePreds( L, v );
                updateEST( L, v, currentTime );
            }
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    }

    verifySolution( L, order, DAG );
    *finalLength = order[L->n-1];
    *finalCost = getLSCost( order, L );
    *costWithFreq = getFreqCost( order, L );

    destroyLS( L );

#ifdef USE_PIPELINING
    // Clean up the pending list
    deletePendingList();
#endif
}

// Compute the cost of a schedule generated by the list scheduler
static int
getLSCost( int order[], LS_type *L )
{
    int i, cost = 0;

    cost = 0;
    for( i = 1; i <= L->n; i++ ) {
        if( L->isSideExit[i] ) { //&& L->costCoeffs[i] > 0 ) {
            cost += order[i - 1] * L->costCoeffs[i];
        }
    }

    return cost;
}

// Compute the cost of a schedule generated by the list scheduler
static double
getFreqCost( int order[], LS_type *L )
{
    int    i;
    double cost;

    cost = 0.0;
    for( i = 1; i <= L->n; i++ ) {
        if( L->isSideExit[i] ) {
            cost += (double)order[i-1] *
                    (double)L->costCoeffs[i] *
                     L->costFreqs[i];
        }
    }

    return( cost );
}

// Find if any instruction already scheduled delays an execution-style instruction.
// Don't count dummy nodes here
BOOL areInstructionsExecuting(LS_type *L, DAG_type *DAG, int order[], int cycle)
{
    int duration, v;

    // Find out if any instructions are executing in the given cycle
    for (v = 1; v <= L->n; v++)
    {
        // Don't count dummy nodes
        if( DAG->instructionType[v - 1] == PSEUDO_INSTR ) {
            continue;
        }

        // Instruction either begins before cycle and ends after
        if (order[v - 1] < cycle && order[v - 1] != -1)
        {
            duration = instructionTimes[DAG->instructionIndex[v - 1]][0];
            if (order[v - 1] + duration - 1 >= cycle)
                return TRUE;
        }
        // Or begins in cycle
        else if (order[v - 1] == cycle)
            return TRUE;
    }

    return FALSE;
}

// Returns true if scheduling bestInstruction in cycle will delay any serial instructions
// in the pending list.
BOOL willDelaySerialNode(int bestInstruction, LS_type *L, DAG_type *DAG, int order[], int cycle)
{
    int duration, v, startCycle = -1;
    LISTNODE* node = getPendingListHead();

    // Need something in the list to be delayed
    if ( isPendingListEmpty() )
        return FALSE;

    // Loop through the list, checking all serial nodes
    while( node != NULL ) {
        if( node->type == INSTRUCTION_TYPE_SERIAL ) {
            // Find the earliest cycle in which the serial instruction can be dispatched
            for (v = 1; v <= L->n; v++)
            {
                if (order[v - 1] < cycle && order[v - 1] != -1)
                {
                    duration = instructionTimes[DAG->instructionIndex[v - 1]][0];
                    startCycle = max(startCycle, order[v - 1] + duration);
                }
            }

            // Now see if we can slip bestInstruction in
            duration = instructionTimes[DAG->instructionIndex[bestInstruction - 1]][0];
            if (cycle + duration - 1 < startCycle)
                return TRUE;
        }

        // Advance through the list
        node = node->next;
    }

    return FALSE;
}

/* =================================================================
 *  Routines for creating and destroying an LS with n variables.
 */

static LS_type *
createLS( int n )
{
    int       i;
    LS_type   *L;

    L = (LS_type *)calloc( (size_t)1, sizeof(LS_type) );

    L->n = n;

    L->varType     = (int *)calloc( (size_t)n+1, sizeof(int) );
    L->maxDelay    = (int *)calloc( (size_t)n+1, sizeof(int) );
    L->EST         = (int *)calloc( (size_t)n+1, sizeof(int) );
    L->nPreds      = (int *)calloc( (size_t)n+1, sizeof(int) );
    L->nSuccs      = (int *)calloc( (size_t)n+1, sizeof(int) );
    L->costCoeffs  = (int *)calloc( (size_t)n+1, sizeof(int) );
    L->costFreqs  = (double *)calloc( (size_t)n+1, sizeof(double) );
    L->dhasy       = (int *)calloc( (size_t)n+1, sizeof(int) );
    L->weight_est  = (int *)calloc( (size_t)n+1, sizeof(int) );
    L->isSideExit  = (BOOL *)calloc( (size_t)n+1, sizeof(BOOL) );
    L->isScheduled = (BOOL *)calloc( (size_t)n+1, sizeof(BOOL) );

    L->latencySucc = (ELEM **) calloc( (size_t)n+1, sizeof(ELEM *) );
    L->latencyPred = (ELEM **) calloc( (size_t)n+1, sizeof(ELEM *) );
    for( i = 1; i <= n; i++ ) {
    L->latencySucc[i] = (ELEM *)NULL;
    L->latencyPred[i] = (ELEM *)NULL;
    }
    
    L->branchDist  = (int **)calloc( (size_t)n+1, sizeof(int *) );
    for( i = 1; i <= n; i++) {
        L->branchDist[i] = (int *)calloc( (size_t)n+1, sizeof(int) );
    }

    L->cp = (int **)calloc( (size_t)L->n+1, sizeof(int *) );
    for( i = 0; i <= L->n; i++ ) {
    L->cp[i] = (int *)calloc( (size_t)L->n+1, sizeof(int) );
    }

    L->pl = (int **)calloc( (size_t)L->n+1, sizeof(int *) );
    for( i = 0; i <= L->n; i++ ) {
    L->pl[i] = (int *)calloc( (size_t)L->n+1, sizeof(int) );
    }

    L->estimates = (int **)calloc( (size_t)L->n+1, sizeof(int *) );
    for( i = 1; i <= L->n; i++ ) {
    L->estimates[i] = (int *)calloc( (size_t)L->n+1, sizeof(int) );
    }

    L->nDescs       = (int *)calloc( (size_t)n+1, sizeof(int) );
    L->distance     = (int *)calloc( (size_t)n+1, sizeof(int) );
    L->eTime        = (int *)calloc( (size_t)n+1, sizeof(int) );
    L->sumLatencies = (int *)calloc( (size_t)n+1, sizeof(int) );
    //L->nOps_s = (int *)calloc( (size_t)N_INSTRUCTION_TYPES+1, sizeof(int) );
    //L->nOps_d = (int *)calloc( (size_t)N_INSTRUCTION_TYPES+1, sizeof(int) );

    return( L );
}

static void
destroyLS( LS_type *L )
{
    int  i;

    free( L->varType );
    free( L->maxDelay );
    free( L->EST );
    free( L->nPreds );
    free( L->nSuccs );
    free( L->costCoeffs );
    free( L->costFreqs );
    free( L->dhasy );
    free( L->weight_est );
    free( L->isSideExit );
    free( L->isScheduled );

    for( i = 1; i <= L->n; i++ ) {
    destroyList( L->latencySucc[i] );
    destroyList( L->latencyPred[i] );
    }
    free( L->latencySucc );
    free( L->latencyPred );

    for( i = 1; i <= L->n; i++ ) {
        free(L->branchDist[i]);
    }
    free(L->branchDist);

    for( i = 0; i <= L->n; i++ ) {
    free( L->cp[i] );
    }
    free( L->cp );

    for( i = 0; i <= L->n; i++ ) {
    free( L->pl[i] );
    }
    free( L->pl );

    for( i = 1; i <= L->n; i++ ) {
    free( L->estimates[i] );
    }
    free( L->estimates );

    free( L->nDescs );
    free( L->distance );
    free( L->eTime );
    free( L->sumLatencies );
    //free( L->nOps_s );
    //free( L->nOps_d );

    free( L );
}

/*
 *  Is the instruction a side exit?
 * 
 *  Used in the side-exit first heuristic and
 *  the DHASY heuristic. For the correctness
 *  of the DHASY heuristic, the final exit is
 *  labeled as a side exit.
 *  Assumes setCostCoeffs has been called.
 */
static void
setSideExits( LS_type *L )
{
    int v;

    for( v = 1; v <= L->n; v++ ) {
    L->isSideExit[v] = FALSE;
    }

    for( v = 1; v < L->n; v++ ) {
    if( L->costCoeffs[v] != -1 ) {
        L->isSideExit[v] = TRUE;
    }
    }
    L->isSideExit[L->n] = TRUE;
}

/*
 *  Set the latency constraints.
 *
 *  latencySucc[w] is a list of all nodes v and latencies l
 *  such that there is a latency constraint between w and v
 *  of the form: lower[w] + l <= lower[v].
 *  latencyPred[w] is a list of all nodes v and latencies l
 *  such that there is a latency constraint between w and v
 *  of the form: upper[w] - l >= upper[v].
 */
static void
setLatencyConstraints( LS_type *L, DAG_type *DAG )
{
    int  v, w, latency;
    ELEM *q;
#ifdef USE_PIPELINING
    int duration;
#endif

    for( v = 0; v < L->n-1; v++ ) {
    for( q = DAG->successors[v]; q != NULL; q = q->next ) {
        w       = q->vertex;
        latency = q->latency;   // latency on edge (v,w)
        /*
         *  Externally, the vertices are named 0, ..., n-1.
         *  Internally, the vertices are named 1, ..., n.
         */
        insertList( &L->latencySucc[v+1], w+1, latency );
        insertList( &L->latencyPred[w+1], v+1, latency );
    }
    }

    // Add a constraint from each node to the sink, with the 
    // latency set to the node's duration. This ensures that
    // no nodes continue executing after the sink's issue
    // cycle.
#ifdef USE_PIPELINING
    for( v = 1; v < L->n; v++ ) {
        duration = instructionTimes[DAG->instructionIndex[v - 1]][0];
        insertList( &L->latencySucc[v], L->n, duration );
        insertList( &L->latencyPred[L->n], v, duration );
    }
#endif
    
    for( v = 1; v <= L->n; v++ ) {
    L->nPreds[v] = length( L->latencyPred[v] );
    L->nSuccs[v] = length( L->latencySucc[v] );
    }
}

/*
 *  Set the earliest start time of the instructions by
 *  propagating the latency constraints.
 */
static void
setEST( LS_type *L )
{
    int  v, w;
    ELEM *p;

    for( v = 1; v <= L->n; v++ ) {
    L->EST[v] = 1;
    }

    for( v = 1; v <= L->n; v++ ) {
    for( p = L->latencySucc[v]; p != NULL; p = p->next ) {
        w = p->vertex;
        if( L->EST[w] < (L->EST[v] + p->latency) ) {
        L->EST[w] =  L->EST[v] + p->latency;
        }
    }
    }
}

/*
 *  Update the earliest start times of the instructions that
 *  are the successors of a recently issued instruction by
 *  propagating the latency constraints.
 */
static void
updateEST( LS_type *L, int v, int issueTime )
{
    int  w;
    ELEM *p;

    for( p = L->latencySucc[v]; p != NULL; p = p->next ) {
        w = p->vertex;
        if( L->EST[w] < (issueTime + p->latency) ) {
            L->EST[w] =  issueTime + p->latency;
        }
    }
}

/*
 *  Update the ready list by checking whether the successors
 *  of a recently issued instruction are now ready to be
 *  scheduled.
 */
static void
updatePreds( LS_type *L, int v )
{
    int  w;
    ELEM *p;

    for( p = L->latencySucc[v]; p != NULL; p = p->next ) {
    w = p->vertex;
    L->nPreds[w]--;
    }
}

/*
 *  Set the maxDelay by propagating the latency constraints.
 */
static void
setMaxDelay( LS_type *L )
{
    int  v, w;
    ELEM *p;

    for( v = L->n; v >= 1; v-- ) {
    L->maxDelay[v] = 1;
    }

    for( v = L->n; v >= 1; v-- ) {
    for( p = L->latencyPred[v]; p != NULL; p = p->next ) {
        w = p->vertex;
        if( L->maxDelay[w] < (L->maxDelay[v] + p->latency) ) {
        L->maxDelay[w] =  L->maxDelay[v] + p->latency;
        }
    }
    }
}

/*
 *  Set the path lengths by propagating the latency constraints.
 */
static void
setPLdistances( LS_type *L )
{
    int  s, v, w;
    ELEM *p;

    for( v = 1; v <= L->n; v++ ) {
    for( w = 1; w <= L->n; w++ ) {
        L->pl[v][w] = -INFINITY;
    }
    L->pl[v][v] = 0;
    }

    for( s = 1; s <= L->n; s++ )
    for( v = s; v <= L->n; v++ ) {
        for( p = L->latencySucc[v]; p != NULL; p = p->next ) {
        w = p->vertex;
        if( L->pl[s][w] < (L->pl[s][v] + 1) ) {
        L->pl[s][w] =  L->pl[s][v] + 1;
        }
    }
    }
}

/*
 *  Set the critical path distances by propagating the latency constraints.
 */
static void
setCPdistances( LS_type *L )
{
    int  s, v, w;
    ELEM *p;

    for( v = 1; v <= L->n; v++ ) {
    for( w = 1; w <= L->n; w++ ) {
        L->cp[v][w] = -INFINITY;
    }
    L->cp[v][v] = 0;
    }

    for( s = 1; s <= L->n; s++ )
    for( v = s; v <= L->n; v++ ) {
        for( p = L->latencySucc[v]; p != NULL; p = p->next ) {
        w = p->vertex;
        if( L->cp[s][w] < (L->cp[s][v] + p->latency) ) {
        L->cp[s][w] = (L->cp[s][v] + p->latency);
        }
    }
    }
}

/*
 *  Set the execution times.
 */
static void
setETime( LS_type *L )
{
    int  v, maxLatency;
    ELEM *p;

    for( v = 1; v <= L->n; v++ ) {
    maxLatency = 0;
    for( p = L->latencySucc[v]; p != NULL; p = p->next ) {
        if( maxLatency < p->latency ) {
            maxLatency = p->latency;
        }
    }
    L->eTime[v] = maxLatency;
    }
}

/*
 *  Set the sum of latencies to successors.
 */
static void
setSumLatencies( LS_type *L )
{
    int  v, sum;
    ELEM *p;

    for( v = 1; v <= L->n; v++ ) {
    sum = 0;
    for( p = L->latencySucc[v]; p != NULL; p = p->next ) {
        sum += p->latency;
    }
    L->sumLatencies[v] = sum;
    }
}

/*
 *  Set the number of descendants of each node.
 */
static void
setDescendants( LS_type *L )
{
    int  v, w;

    for( v = 1; v <= L->n; v++ ) {
    L->nDescs[v] = 0;
    for( w = v+1; w <= L->n; w++ ) if( L->cp[v][w] >= 0 ) {
        L->nDescs[v]++;
    }
    }
}

static void     
setDistances( LS_type * L)
{
  int * branchIds;
  int i,j;
  ELEM *p;
  int exitCount;
  int count;
  int currentDistance;
  int predDistance;
  int currentId;
  /* count the number of side exits*/
  exitCount = 0;
  for(i = 1; i <= L->n; i++)
    if(L->isSideExit[i])
      exitCount++;

  /* allocate enough space to store all the side exit ids*/
  branchIds = (int *)calloc((size_t)exitCount, sizeof(int));

  /* copy all of the ids of the side exits into the array*/
  count = 0;
  for(i = L->n; i > 0; i--)
    {
      if(L->isSideExit[i])
    {
      branchIds[count] = i;
      count++;
    }
    }
  /* initialize each vertexes distance array for each branch */

  for(i = 1; i <= L->n; i++)
    for(j = 0; j < exitCount; j++)
      L->branchDist[i][branchIds[j]] = -1;

  /* for each side exit calculate the distance to the other exits */
  for(i = 0; i < exitCount; i++)
    {
      currentId = branchIds[i];
      L->branchDist[currentId][currentId] = 0;
      for(j = L->n; j > 0; j--)
    {
      currentDistance = L->branchDist[j][currentId];
      if(currentDistance >= 0)
        {
          for( p = L->latencyPred[j]; p != NULL; p = p->next )
        {
          predDistance = L->branchDist[p->vertex][currentId];
          if(predDistance < currentDistance + p->latency)
            L->branchDist[p->vertex][currentId] = currentDistance + p->latency;
        }
        }
    }
    }

  /* free up the branch ids array */
  free( branchIds );
}

static void     
printCostCoeffs( LS_type *L )
{
    int    i;

    for( i = 1; i < L->n; i++ ) {
    if(L->costCoeffs[i] != -1) {
        printf("%d x_%d + ", L->costCoeffs[i], i);
    }
    }
    printf("%d x_%d\n", L->costCoeffs[L->n], L->n);

}

static void     
setCostCoeffs( LS_type *L, DAG_type *DAG)
{
    int    i;
    double freq_fallthrough;
    double percent, coeff;

    /*
     *  Determine the probability of exiting at each side exit.
     */
    percent = 1.0;
    // -1 since the last node cannot be a side exit.
    for( i = 0; i < DAG->nInstructions-1; i++ ) {
        if( DAG->percentage[i] ) {
            coeff = percent * ((double)DAG->percentage[i]);
#ifdef WIN32
        // No round method in the math library under Windows
            L->costCoeffs[i+1] = (int)coeff;
#else
        L->costCoeffs[i+1] = (int)rint(coeff);
#endif
        L->costFreqs[i+1] = DAG->f1[i] + DAG->f2[i];
        if( L->costFreqs[i+1] == 0.0 ) {
            L->costFreqs[i+1] = 1.0;
        }
        if( L->costCoeffs[i+1] == 0 ) {
        L->costCoeffs[i+1] = -1;
        }
            percent *= (1.0 - ((double)DAG->percentage[i])/100.0);
        }
    else {
        L->costCoeffs[i+1] = -1;
        L->costFreqs[i+1] = 0.0;
    }
    }
    /*
     *  Determine the probability that we fall through to the sink node.
     */
    L->costCoeffs[L->n] = 100;
    //for( i = 0; i <= DAG->nInstructions-1; i++ ) {
    for( i = 0; i < DAG->nInstructions-1; i++ ) {
    if(L->costCoeffs[i+1] != -1) {
        L->costCoeffs[L->n] -= L->costCoeffs[i+1];
    }
    }
    freq_fallthrough = 1.0E100;
    for( i = 0; i < DAG->nInstructions-1; i++ ) {
        if( DAG->percentage[i] && DAG->f2[i] > 0.0 ) {
        if( freq_fallthrough > DAG->f2[i] ) {
            freq_fallthrough = DAG->f2[i];
        }
    }
    }
    if( freq_fallthrough == 0.0 || freq_fallthrough == 1.0E100 ) {
    freq_fallthrough = 1.0;
    }
    L->costFreqs[L->n] = freq_fallthrough;
    /*
     *  Rarely the fall through probability is zero or negative
     *  (if rounding goes against us). If so, set it to the
     *  smallest value.
     */
    if( L->costCoeffs[L->n] <= 0 ) {
        L->costCoeffs[L->n] = 1;
    }
}

static void     
calcDhasy1( LS_type *L)
{
  int i,j;
  int cost;
  int lastbranch;
  int branch_dist;
  int distance;

  /*find the last side exit and store the id*/
  lastbranch = L->n;
  for(i = L->n; i > 0; i--)
    if(L->isSideExit[i])
      {
    lastbranch = i;
    break;
      }

  for(i = 1; i <= L->n; i++)
    {
      cost = 0;
      for(j = 1; j <= L->n; j++)
    if(L->isSideExit[j] && L->costCoeffs[j] > 0)
      {
        branch_dist = L->branchDist[j][lastbranch];
        distance = L->branchDist[i][j];
        if(distance >= 0)
          cost += L->costCoeffs[j] * (branch_dist + 1 + distance);
      }
      L->dhasy[i] = cost;
    }
}

static void     
calcDhasy2( LS_type *L )
{
    int i, j, cost;

    for( i = 1; i <= L->n; i++ ) {
    cost = 0;
    for( j = i; j <= L->n; j++ ) if( L->isSideExit[j] ) {
        if( L->cp[i][j] >= 0 ) {
        cost += L->costCoeffs[j] *
            (L->cp[1][L->n] + 1 - (L->cp[1][j] - L->cp[i][j]));
        }
    }
    L->dhasy[i] = cost;
    }
}

static void     
calcDhasy3( LS_type *L )
{
    int i, j, cost;

    for( i = 1; i <= L->n; i++ ) {
    cost = 0;
    for( j = i; j <= L->n; j++ ) if( L->isSideExit[j] ) {
        if( L->cp[i][j] >= 0 ) {
        cost += L->costCoeffs[j] * (L->cp[i][j] + 1);
        }
    }
    L->dhasy[i] = cost;
    }
}

/*
 *  Determine critical path distances for each vertex.
 *  Assumption: all successors of a vertex have a higher vertex number.
 */
static void
criticalPathDistances( LS_type *L, DAG_type *DAG )
{
    int  s, u, v;
    ELEM *q;

    for( s = 1; s <= L->n; s++ ) {
    for( v = 1; v <= L->n; v++ ) {
        L->cp[s][v] = -INFINITY;
    }
    L->cp[s][s] = 0;
    }

    for( s = 1; s <= L->n; s++ )
    for( u = s; u <= L->n; u++ ) {
        for( q = DAG->successors[u-1]; q != NULL; q = q->next ) {
        v = q->vertex + 1;
        if( L->cp[s][v] < (L->cp[s][u] + q->latency) ) {
        L->cp[s][v] = (L->cp[s][u] + q->latency);
        }
    }
    }
}

#define MIN_N  3

static int
estimateRegion( LS_type *L, int s, int t, BOOL *inRegion )
{
    int      i, j, max_i, max_j, max_a, max_b, w;
    int      distance, t_distance, d;
    int      type, nOpsRegion[N_INSTRUCTION_TYPES];
    int      ms[N_INSTRUCTION_TYPES][MIN_N+1];
    int      ns[N_INSTRUCTION_TYPES][MIN_N+1];
    int      mp[N_INSTRUCTION_TYPES][MIN_N+1];
    int      np[N_INSTRUCTION_TYPES][MIN_N+1];

    for( type = 0; type < N_INSTRUCTION_TYPES; type++ ) {
    for( j = 0; j <= MIN_N; j++ ) {
        ms[type][j] = INFINITY;
        ns[type][j] = 0;
        mp[type][j] = INFINITY;
        np[type][j] = 0;
    }
    nOpsRegion[type] = 0;
    }

    /*
     *  Count the nodes w that are necessarily
     *  between s and t (i.e., on a path from s to t).
     *  Determine the minimum critical path distances.
     */
    for( w = s; w <= t; w++ ) if( inRegion[w] ) {
    type = L->varType[w];
    nOpsRegion[type]++;
    if( L->cp[s][w] <= ms[type][MIN_N] ) {
        i = 0;
        while( L->cp[s][w] > ms[type][i] ) {
        i++;
        }
        if( L->cp[s][w] == ms[type][i] ) {
        for( j = MIN_N; j > i; j-- ) {
            ns[type][j]++;
        }
        }
        else
        if( L->cp[s][w] < ms[type][i] ) {
        for( j = MIN_N; j > i; j-- ) {
            ms[type][j] = ms[type][j-1];
            ns[type][j] = ns[type][j-1]+1;
        }
        ms[type][i] = L->cp[s][w];
        }
    }

    if( L->cp[w][t] <= mp[type][MIN_N] ) {
        i = 0;
        while( L->cp[w][t] > mp[type][i] ) {
        i++;
        }
        if( L->cp[w][t] == mp[type][i] ) {
        for( j = MIN_N; j > i; j-- ) {
            np[type][j]++;
        }
        }
        else
        if( L->cp[w][t] < mp[type][i] ) {
        for( j = MIN_N; j > i; j-- ) {
            mp[type][j] = mp[type][j-1];
            np[type][j] = np[type][j-1]+1;
        }
        mp[type][i] = L->cp[w][t];
        }
    }
    }

    /*
     *  Lower bound on the distance between s and t.
     */
    distance = 0;
    for( type = 0; type < N_INSTRUCTION_TYPES; type++ )  {
    if( nOpsRegion[type] != 0 ) {

        max_a = 0;
        max_i = 0;
        for( i = 0; i <= MIN_N; i++ ) if( ms[type][i] != INFINITY ) {
        d = ms[type][i] + (int)ceil(((double)nOpsRegion[type]-ns[type][i]) /
                        ((double)nFunctionalUnits( type )));
        if( max_a < d ) {
            max_a = d;
            max_i = i;
        }
        }

        max_b = 0;
        max_j = 0;
        for( i = 0; i <= MIN_N; i++ ) if( mp[type][i] != INFINITY && nOpsRegion[type]-ns[type][max_i]-np[type][i]>0 ) {
        d = mp[type][i] + (int)ceil(((double)nOpsRegion[type]-np[type][i]) /
                        ((double)nFunctionalUnits( type )));

        if( max_b < d ) {
            max_b = d;
            max_j = i;
        }
        }

        t_distance =
        ms[type][max_i] +
        (int)ceil(((double)nOpsRegion[type]-ns[type][max_i]-np[type][max_j]) /
             ((double)nFunctionalUnits( type ))) +
        mp[type][max_j] +
        - 1;
        if( t_distance > distance ) distance = t_distance;
    }
    }

    return( distance );
}

/*
 *  Determine estimated distances for each vertex.
 *  Assumption: all successors of a vertex have a higher vertex number.
 */
static void
setEstimatedDistances( LS_type *L, DAG_type *DAG )
{
    int  s, t, w;
    BOOL *inRegion;

    inRegion = (BOOL *)calloc( (size_t)L->n+1, sizeof(BOOL) );

    for( s = 1; s <= L->n; s++ ) {
    for( t = 1; t <= L->n; t++ ) {
        L->estimates[s][t] = -INFINITY;
    }
    L->estimates[s][s] = 0;
    }

    for( s =   1; s <= L->n; s++ )
    for( t = s+1; t <= L->n; t++ ) if( L->cp[s][t] >= 0 ) {

    for( w = 1; w <= L->n; w++ ) {
        inRegion[w] = FALSE;
    }
    inRegion[s] = TRUE;
    inRegion[t] = TRUE;
    for( w = s+1; w < t; w++ ) {
        if( (L->cp[s][w] >= 0) && (L->cp[w][t] >= 0) ) {
        inRegion[w] = TRUE;
        }
    }

    L->estimates[s][t] = estimateRegion( L, s, t, inRegion );
    }

    free( inRegion);
}

static void     
calcWeightedEstimate( LS_type *L )
{
    int i, j, cost;

    for( i = 1; i <= L->n; i++ ) {
    cost = 0;
    for( j = i; j <= L->n; j++ ) if( L->isSideExit[j] ) {
        if( L->estimates[i][j] >= 0 ) {
        cost += L->costCoeffs[j] *
            (L->estimates[1][L->n] + 1 - (L->estimates[1][j] - L->estimates[i][j]));
        }
    }
    L->weight_est[i] = cost;
    }
}

/* =================================================================
 *  Routines for printing/debugging.
 */

static void
printLS( LS_type *L )
{
    int  v;
    ELEM *p;

    printf( "\n\tn: %d\n", L->n );
    for( v = 1; v <= L->n; v++ ) {
    printf( "\t%3d: est = %3d maxDelay = %3d nPreds = %3d isSched = %d; type=%d\n",
        v,
        L->EST[v],
        L->maxDelay[v],
        L->nPreds[v],
        L->isScheduled[v],
        L->varType[v] );
    }
    printf( "\n" );

    printf( "Successors:\n" );
    for( v = 1; v <= L->n; v++ ) {
    printf( "%d", v );
    for( p = L->latencySucc[v]; p != NULL; p = p->next ) {
        printf( " %d:%d", p->vertex, p->latency );
    }
    printf( "\n" );
    }
    printf( "\n" );

    printf( "Predecessors:\n" );
    for( v = 1; v <= L->n; v++ ) {
    printf( "%d", v );
    for( p = L->latencyPred[v]; p != NULL; p = p->next ) {
        printf( " %d:%d", p->vertex, p->latency );
    }
    printf( "\n" );
    }
    printf( "\n" );
}

static void
verifySolution( LS_type *L, int order[], DAG_type *DAG )
{
    int  t, v, m, type, status, *count;
    ELEM *p;
#ifdef USE_PIPELINING
    int duration, room;
#endif

    m = 0;
    for( v = 1; v <= L->n; v++ ) {
    if( m < order[v-1] ) m = order[v-1];
    }
    count = (int *)calloc( (size_t)m+1, sizeof(int) );

    status = 1;
    for( v = 1; v <= L->n; v++ ) {
    status = status && (order[v-1] > 0);
    }
    if( !status ) {
    printf( "Verifying List Scheduler solution:\n" );
    printf( "\tEach instruction issued not in range [1, %d].\n", m );
    exit( 6 );
    }

    for( type = 0; type < N_INSTRUCTION_TYPES; type++ ) {
    for( t = 1; t <= m; t++ ) {
        count[t] = 0;
    }
    for( v = 1; v <= L->n; v++ ) {
        if( L->varType[v] == type )
        count[order[v-1]]++;
    }
    status = 1;
    for( t = 1; t <= m; t++ ) {
        status = status && (count[t] <= nFunctionalUnits( type ));
    }
    if( !status ) {
        printf( "Verifying List Scheduler solution:\n" );
        printf( "\tIssue-width constraint on type %d not satisfied.\n", type );
        exit( 6 );
    }
    }

    /* Verify that the overall issue width isn't violated */
    for( t = 1; t <= m; t++ ) {
        count[t] = 0;
    }
    for( v = 1; v <= L->n; v++ ) {
        // Don't count pseudo-instructions
        if( L->varType[v] != PSEUDO_INSTR )
            count[order[v-1]]++;
    }
    status = 1;
    for( t = 1; t <= m; t++ ) {
        status = status && (count[t] <= getIssueWidth());
    }
    if ( !status) {
        printf( "Verifying List Scheduler solution:\n" );
        printf( "\tOverall issue-width constraint not satisfied.\n" );
        //exit( 6 );
    }
    
    status = 1;
    for( v = 1; v <= L->n; v++ ) {
    for( p = L->latencySucc[v]; p != NULL; p = p->next ) {
        status = status &&
            ((order[v-1] + p->latency) <= order[p->vertex-1]);
    }
    }
    if( !status ) {
    printf( "Verifying List Scheduler solution:\n" );
    printf( "\tSuccessor constraints not satisfied.\n" );
    exit( 6 );
    }

    status = 1;
    for( v = 1; v <= L->n; v++ ) {
    for( p = L->latencyPred[v]; p != NULL; p = p->next ) {
        status = status &&
            ((order[p->vertex-1] + p->latency) <= order[v-1]);
    }
    }
    if( !status ) {
    printf( "Verifying List Scheduler solution:\n" );
    printf( "\tPredecessor constraints not satisfied.\n" );
    exit( 6 );
    }

#ifdef USE_PIPELINING
    // Step through time, making sure that there's enough resources at each cycle
    for( type = 0; type < N_INSTRUCTION_TYPES; type++ ) {
        for( t = 1; t <= m; t++ ) {
            room = nFunctionalUnits( type );
            for( v = 1; v <= L->n; v++ ) {
                if( L->varType[v] == type ) {
                    if( order[v - 1] == t ) {
                        --room;
                    }
                    else if( order[v - 1] < t ) {
                        duration = instructionTimes[DAG->instructionIndex[v - 1]][0];
                        if( order[v - 1] + duration - 1 >= t ) {
                            --room;
                        }
                    }

                    if( room < 0 ) {
                        printf( "Verifying List Scheduler solution:\n" );
                        printf( "\tPipelining constraint on type %d not satisfied in cycle %d.\n", type, t );
                        exit( 6 );
                    }
                }
            }
        }
    }

#endif

#ifdef USE_EXECUTION_SERIALIZATION
    for( v = 1; v <= L->n; v++ ) {
        // If this instruction is execution-style, check it
        if( isExecution( DAG->instructionIndex[v - 1] ) ) {
            for( t = 1; t <= L->n; t++) {
                if (t == v || DAG->instructionType[t - 1] == PSEUDO_INSTR) {
                    continue;
                }

                // Instruction t can't start before v and end after
                if ( order[v - 1] > order[t - 1] ) {
                    duration = instructionTimes[DAG->instructionIndex[t - 1]][0];
                    if ( order[t - 1] + duration - 1 > order[v - 1] ) {
                        printf( "Verifying List Scheduler solution:\n" );
                        printf( "\tExecution-style instruction wasn't the last incomplete instruction when it began execution.\n" );
                        exit( 6 );
                    }
                }

                // Instruction also can't begin in the same cycle as v
                if( order[v - 1] == order[t - 1] ) {
                    printf( "Verifying List Scheduler solution:\n" );
                    printf( "\tMore than one instruction issued in a cycle when an execution-style instruction was issued\n" );
                    exit( 6 );
                }
            }
        }
    }
#endif
    
    free( count );
}
