
#include <assert.h>
#include <string.h>
#include "global.h"
#include "processor.h"

/* =================================================================
 *  Routines for answering queries about the processor model
 */

/*
 *  Number of functional units of each instruction type.
 *  The order of the values corresponds to the order of
 *  the instruction types below.
 */
static int n_func_units[N_FUNCTIONAL_UNIT_TYPES];


/*
 * Issue width for the machine.
 * Need not be equal to the number of functional units
 */
static int issueWidth;

/*
 * Maximum latency of any instruction
 */
static int maxLatency;


/*
 *  Names of instructions types as they appear in the input files.
 *  Internally the names are 0, 1, ...
 *  Assumption: DUMMY instruction is always listed last since this is
 *  assumed by propagator (the resource constraint is not propagated on
 *  the last instruction type). Even if there are no DUMMY instructions,
 *  they should be included.
 */
static char *instruction_type[N_INSTRUCTION_TYPES] = {
    "FXU",	// integer instructions
    "FPU",	// floating point instructions
    "LSU",	// load or store instruction
    "BPU",	// branching instructions
	"SFX",  // simple integer instructions
	"CFX",  // complex integer instructions
    "DUMMY"	// compiler directive pseudo-instructions
};


/*
 * Mapping between instruction type and functional unit
 */
static int unitMapping[N_INSTRUCTION_TYPES];

/*
 * Preference for instruction types in list scheduling.
 */
static int instructionTypePreference[N_INSTRUCTION_TYPES];


/*
 * Get the instruction type for a particular instruction
 */
int
getInstructionType( char *line )
{
    int i;
	char temp[1024];

	/* Ensure all instructions execute on only one unit type */
    if( strcmp( line, "FXU+BPU" ) == 0 ) {
		strcpy( temp, "FXU" );
    }
    else if( strcmp( line, "FPU+BPU" ) == 0 ) {
		strcpy( temp, "FPU" );
    }
	else if( strcmp( line, "SFX+BPU" ) == 0 ) {
		strcpy( temp, "SFX" );
	}
	else if( strcmp( line, "CFX+BPU" ) == 0 ) {
		strcpy( temp, "CFX" );
	}
	else {
		strcpy( temp, line );
	}

	/* Find the instruction type */
    for( i = 0; i < N_INSTRUCTION_TYPES; i++ ) {
		if( strcmp( temp, instruction_type[i] ) == 0 ) {
			break;
		}
    }
    
	if( i == N_INSTRUCTION_TYPES && strcmp( line, "ERROR") != 0 ) {
		fprintf( stderr, "Instruction type not found: %s\n", temp );
		exit( 6 );
    }
	else if( i == N_INSTRUCTION_TYPES ) {
		return -1;
	}

	return getInstructionUnitMapping( i );
}

void
setNumFunctionalUnits( int type, int nUnits )
{
	assert( type >= 0 && type < N_FUNCTIONAL_UNIT_TYPES );
	assert( nUnits >= 0 );

    n_func_units[type] = nUnits;
}

int
nFunctionalUnits( int type )
{
	assert( type >= 0 && type < N_FUNCTIONAL_UNIT_TYPES );

    return( n_func_units[type] );
}

// Returns max(nFunctionalUnits(t)) over all types t
int getMaxUnitsOfAnyType()
{
	int largest = -1, i = 0;
	for( i = 0; i < N_FUNCTIONAL_UNIT_TYPES; i++ ) {
		if( nFunctionalUnits(i) > largest &&
			i != PSEUDO_UNIT ) {
			largest = i;
		}
	}

	return largest;
}

void
setIssueWidth( int width )
{
	assert( issueWidth >= 0 );

	issueWidth = width;
}

int
getIssueWidth()
{
#ifdef USE_ISSUE_WIDTH
	return issueWidth;
#else
	int i, sum = 0;

	for (i = 0; i < N_FUNCTIONAL_UNIT_TYPES; i++)
		sum += n_func_units[i];
	return sum;
#endif
}


void
setMaxLatency( int latency )
{
	assert(latency >= 0);

	maxLatency = latency;
}

int
getMaxLatency()
{
	return maxLatency;
}

/*
 * Get the functional unit set to execute a given instructino type
 */
int
getInstructionUnitMapping( int instructionType )
{
	assert( instructionType >= 0 && instructionType < N_INSTRUCTION_TYPES );

	return unitMapping[instructionType];
}

/* 
 * Define which functional unit executes a given instruction type
 */
void
setInstructionUnitMapping( int instructionType, int unitType )
{
	assert( instructionType >= 0 && instructionType < N_INSTRUCTION_TYPES );
	assert( unitType >= 0 && unitType < N_FUNCTIONAL_UNIT_TYPES );

	unitMapping[instructionType] = unitType;
}

/*
 * Set the preference for instruction types for use in list scheduling
 */
int
getInstructionTypePreference( int rank )
{
	assert(rank >= 0 && rank < N_INSTRUCTION_TYPES );

	return instructionTypePreference[rank];
}

void
setInstructionTypePreference( int rank, int instructionType )
{
	assert( rank >= 0 && rank < N_INSTRUCTION_TYPES );
	assert( ( instructionType >= 0 && instructionType < N_INSTRUCTION_TYPES ) ||
		instructionType == INSTR_DONT_CARE );

	instructionTypePreference[rank] = instructionType;
}
