/********************************************************************
* settings.c                                                        *
*********************************************************************
* Written by Mike Chase                                             *
*********************************************************************
* Purpose: Providing a way to easily specify common data for a      *
* scheduling experiment in a single experiment definition file      *
* instead of running different experiments by a combination of      *
* command line trickery and recompiling the code with different     *
* parts of the code commented out.                                  *
********************************************************************/

#include <string.h>
#include "global.h"
#include "csp.h"
#include "helper.h"
#include "processor.h"
#include "settings.h"
#include "list.h"
#include "dag.h"
#include "ls.h"
#include "heuristic.h"

BOOL requestedLS, requestedOptimal, requestedPrintSchedule, schedulingSuperblocks;
BOOL lsForward, speculation;
char baseDataPath[2049];

void setDefaultSettings();
int  findSetting( char *settingName );
int  getUnitByString( char *unitName );
int  getHeuristicByString( char *heuristic );
int  getInstrTypeByString( char *instrTypeName );
BOOL addPref( char *instrTypeName, int rank );

/*
 * Setting names
 */
char *settingNames[] = {
	"ListScheduler",
    "OptimalScheduler",
    "PrintSchedule",
    "MaxLatency",
    "IssueWidth",
    "NumIntFunctionalUnits",
    "NumFloatFunctionalUnits",
    "NumLoadStoreFunctionalUnits",
    "NumBranchFunctionalUnits",
    "IntMapping",
    "FloatMapping",
    "LoadStoreMapping",
    "BranchMapping",
    "InstructionPreference",
    "BaseDataPath",
    "TimeoutLevel1",
    "TimeoutLevel2",
    "TimeoutLevel3",
    "MaxRegionSize",
    "MaxRestartSize",
	"NumSimpleIntUnits",
	"NumComplexIntUnits",
	"SimpleIntMapping",
	"ComplexIntMapping",
	"LSHeuristic",
	"OptHeuristic",
	"SuperBlocks",
	"LSForward",
	"Speculation"
};


/*
 * Settings variables
 */
#define NUM_SETTINGS            29

#define SETTINGS_LS             0
#define SETTINGS_OPT            1
#define SETTINGS_PRINT          2
#define SETTINGS_MAXLATENCY     3
#define SETTINGS_ISSUEWIDTH     4
#define SETTINGS_N_INT_UNITS    5
#define SETTINGS_N_FLOAT_UNITS  6
#define SETTINGS_N_LDST_UNITS   7
#define SETTINGS_N_BRANCH_UNITS 8
#define SETTINGS_INT_MAPPING    9
#define SETTINGS_FLOAT_MAPPING  10
#define SETTINGS_LDST_MAPPING   11
#define SETTINGS_BRANCH_MAPPING 12
#define SETTINGS_INSTR_PREF     13
#define SETTINGS_DATA_PATH      14
#define SETTINGS_TIMEOUT_L1     15
#define SETTINGS_TIMEOUT_L2     16
#define SETTINGS_TIMEOUT_L3     17
#define SETTINGS_MAX_RGN_SIZE   18
#define SETTINGS_MAX_RSTRT_SIZE 19
#define SETTINGS_N_SFX_UNITS    20
#define SETTINGS_N_CFX_UNITS    21
#define SETTINGS_SFX_MAPPING    22
#define SETTINGS_CFX_MAPPING    23
#define SETTINGS_LS_HEURISTIC   24
#define SETTINGS_OPT_HEURISTIC  25
#define SETTINGS_SUPERBLOCKS    26
#define SETTINGS_LS_FORWARD     27
#define SETTINGS_SPECULATE      28

/*
 * Instruction type strings
 */
char *instrTypeStrings[] = {
	"INT",
	"FLOAT",
	"LDST",
	"BRANCH",
	"SFX",
	"CFX",
	"PSEUDO", /* Probably unused in settings files */
	"DONTCARE"
};
#define DONTCARE_TYPE_STRING 7

/*
 * Functional unit type strings
 */
char *unitTypeStrings[] = {
	"INT",
	"FLOAT",
	"LDST",
	"BRANCH",
	"SFX",
	"CFX",
	"PSEUDO" /* Probably unused in settings files */
};

/*
 * Heuristic name strings
 */
char *heuristicNameStrings[] = {
	"CP",
	"DT",
	"SP",
	"SP_2",
	"SP_3",
	"EST",
	"BEST",
	"SIDEEXITFIRST",
	"DHASY1",
	"DHASY2",
	"DHASY3",
	"WEIGHTED_ESTIMATE"
};


/*
 * Read a settings file and initialize appropriate settings
 *
 * Returns FALSE if an error occurs
 */
BOOL readSettingsFile(char* filename)
{
	FILE *fp;
	char temp[1025], separators[] = " ,\t\n";
	char *token;
	int setting, pos, mapping, heuristic;

	setDefaultSettings();

	fp = fopen(filename, "rt");
	if( NULL == fp ) {
		printf( "Couldn't open settings file %s\n", filename );
		return FALSE;
	}

	/* Read the file line by line */
	while( !feof(fp) ) {
		fgets( temp, 1024, fp );
		
		/* Parse the current input line */
		token = strtok( temp, separators );
		if ( NULL != token ) {
			/* Check for a comment */
			if( token[0] == '%' ) {
				continue;
			}

			/* Token is a variable or is bad data */
			setting = findSetting( token );
			if( -1 == setting ) {
				continue; /* Ignore the line */
			}

			/* Have a setting; this should be = */
			token = strtok( NULL, separators );
			if( 0 != strcmp( token, "=" ) ) {
				printf( "Expected = following variable %s\n", settingNames[setting] );
				continue;
			}

			/* Now, have one or more values, depending on variable type */
			token = strtok( NULL, separators );
			if( NULL == token ) {
				printf( "Expected values following %s =\n", settingNames[setting] );
				continue;
			}

			/* Grab the setting and assign it */
			switch( setting ) {
				case SETTINGS_LS: {
					requestedLS = atoi( token );
					break;
				}
				case SETTINGS_OPT: {
					requestedOptimal = atoi( token );
					break;
				}				
				case SETTINGS_PRINT: {
					requestedPrintSchedule = atoi( token );
					break;
				}				
				case SETTINGS_MAXLATENCY: {
					setMaxLatency( atoi( token ) );
					break;
				}				
				case SETTINGS_ISSUEWIDTH: {
					setIssueWidth( atoi( token ) );
					break;
				}				
				case SETTINGS_N_INT_UNITS: {
					setNumFunctionalUnits( INT_UNIT, atoi( token ) );
					break;
				}
				case SETTINGS_N_FLOAT_UNITS: {
					setNumFunctionalUnits( FLOAT_UNIT, atoi( token ) );
					break;
				}				
				case SETTINGS_N_LDST_UNITS: {
					setNumFunctionalUnits( LDST_UNIT, atoi( token ) );
					break;
				}				
				case SETTINGS_N_BRANCH_UNITS: {
					setNumFunctionalUnits( BRANCH_UNIT, atoi( token ) );
					break;
				}				
				case SETTINGS_INT_MAPPING: {
					mapping = getUnitByString( token );
					if ( mapping >= 0 && mapping < N_FUNCTIONAL_UNIT_TYPES ) {
						setInstructionUnitMapping( INTEGER_INSTR, mapping );
					}
					else {
						printf( "Unknown functional unit %s\n", token );
						continue;
					}
					break;
				}				
				case SETTINGS_FLOAT_MAPPING: {
					mapping = getUnitByString( token );
					if ( mapping >= 0 && mapping < N_FUNCTIONAL_UNIT_TYPES ) {
						setInstructionUnitMapping( FLOAT_INSTR, mapping );
					}
					else {
						printf( "Unknown functional unit %s\n", token );
						continue;
					}
					break;
				}				
				case SETTINGS_LDST_MAPPING: {
					mapping = getUnitByString( token );
					if ( mapping >= 0 && mapping < N_FUNCTIONAL_UNIT_TYPES ) {
						setInstructionUnitMapping( LDST_INSTR, mapping );
					}
					else {
						printf( "Unknown functional unit %s\n", token );
						continue;
					}
					break;
				}				
				case SETTINGS_BRANCH_MAPPING: {
					mapping = getUnitByString( token );
					if ( mapping >= 0 && mapping < N_FUNCTIONAL_UNIT_TYPES ) {
						setInstructionUnitMapping( BRANCH_INSTR, mapping );
					}
					else {
						printf( "Unknown functional unit %s\n", token );
						continue;
					}
					break;
				}				
				case SETTINGS_INSTR_PREF: {
					pos = 0;
					while( NULL != token && addPref( token, pos ) ) {
						++pos;
						token = strtok( NULL, separators );
					};
					break;
				}				
				case SETTINGS_DATA_PATH: {
					strcpy( baseDataPath, token );
					break;
				}
				case SETTINGS_TIMEOUT_L1: {
					setCSPTimeout( TIME_LIMIT_FIRST, atof( token ) );
					break;
				}
				case SETTINGS_TIMEOUT_L2: {
					setCSPTimeout( TIME_LIMIT_SECOND, atof( token ) );
					break;
				}
				case SETTINGS_TIMEOUT_L3: {
					setCSPTimeout( TIME_LIMIT_TOTAL, atof( token ) );
					break;
				}
				case SETTINGS_MAX_RGN_SIZE: {
					setMaxRegionSize( atoi( token ) );
					break;
				}
				case SETTINGS_MAX_RSTRT_SIZE: {
					setMaxRestartSize( atoi( token ) );
					break;
				}
				case SETTINGS_N_SFX_UNITS: {
					setNumFunctionalUnits( SFX_UNIT, atoi( token ) );
					break;
				}
				case SETTINGS_N_CFX_UNITS: {
					setNumFunctionalUnits( CFX_UNIT, atoi( token ) );
					break;
				}
				case SETTINGS_SFX_MAPPING: {
					mapping = getUnitByString( token );
					if ( mapping >= 0 && mapping < N_FUNCTIONAL_UNIT_TYPES ) {
						setInstructionUnitMapping( SIMPLEINT_INSTR, mapping );
					}
					else {
						printf( "Unknown functional unit %s\n", token );
						continue;
					}
					break;
				}
				case SETTINGS_CFX_MAPPING: {
					mapping = getUnitByString( token );
					if ( mapping >= 0 && mapping < N_FUNCTIONAL_UNIT_TYPES ) {
						setInstructionUnitMapping( COMPLEXINT_INSTR, mapping );
					}
					else {
						printf( "Unknown functional unit %s\n", token );
						continue;
					}
					break;
				}
				case SETTINGS_LS_HEURISTIC: {
					heuristic = getHeuristicByString( token );
					if ( heuristic == HEURISTIC_BEST ) {
						printf( "Can't use heuristic BEST for the list scheduler\n" );
						continue;
					}
					else if ( heuristic >= 0 && heuristic < HEURISTIC_NUM_HEURISTICS ) {
						setLSHeuristic( heuristic );
					}
					else {
						printf( "Unknown heuristic %s\n", token );
						continue;
					}
					break;
				}
				case SETTINGS_OPT_HEURISTIC: {
					heuristic = getHeuristicByString( token );
					if ( heuristic >= 0 && heuristic < HEURISTIC_NUM_HEURISTICS ) {
						setOptHeuristic( heuristic );
					}
					else {
						printf( "Unknown heuristic %s\n", token );
						continue;
					}
					break;
				}
				case SETTINGS_SUPERBLOCKS: {
					schedulingSuperblocks = atoi( token );
					break;
				}
				case SETTINGS_LS_FORWARD: {
					lsForward = atoi( token );
					break;
				}
				case SETTINGS_SPECULATE: {
					speculation = atoi(token);
					break;
				}
			}
		}
	};

	return TRUE;
}

/*
 * See if a given string is a setting
 */
int
findSetting( char *settingName )
{
	int i;

	for( i = 0; i < NUM_SETTINGS; i++ ) {
		if( 0 == stricmp( settingNames[i], settingName ) ) {
			return i;
		}
	}

	return -1;
}

/*
 * Find the ID for a given functional unit name
 */
int
getUnitByString( char *unitName )
{
	int i;

	for( i = 0; i < N_FUNCTIONAL_UNIT_TYPES; i++ ) {
		if( 0 == stricmp( unitTypeStrings[i], unitName ) ) {
			return i;
		}
	}

	return -1;
}

/*
 * Find the ID for a given heuristic name
 */
int
getHeuristicByString( char *heuristicName )
{
	int i;

	for( i = 0; i < HEURISTIC_NUM_HEURISTICS; i++ ) {
		if( 0 == stricmp( heuristicNameStrings[i], heuristicName ) ) {
			return i;
		}
	}

	return -1;
}

/*
 * Find the ID for a given instruction type name
 */
int
getInstrTypeByString( char *instrTypeName )
{
	int i;

	for( i = 0; i < N_INSTRUCTION_TYPES; i++ ) {
		if( 0 == stricmp( instrTypeStrings[i], instrTypeName ) ) {
			return i;
		}
	}

	/* Explicitly check for the don't care value */
	if( 0 == stricmp( instrTypeStrings[DONTCARE_TYPE_STRING], instrTypeName ) ) {
		return INSTR_DONT_CARE;
	}

	return -1;
}

/*
 * Try to add a preference for a given instructional type
 */
BOOL
addPref( char *instrTypeName, int rank )
{
	int typeID;

	typeID = getInstrTypeByString( instrTypeName );
	if( typeID < 0 )
		return FALSE;

	setInstructionTypePreference( rank, typeID );
	return TRUE;
}


	
/*
 * Set up default settings in case some settings aren't specified in the file
 */
void setDefaultSettings()
{
	/* Operations */
	requestedLS = TRUE;
	requestedOptimal = FALSE;
	requestedPrintSchedule = FALSE;
	schedulingSuperblocks = FALSE;
	lsForward = TRUE;

	/* General Architecture */
	setMaxLatency(50);
	setIssueWidth(4);
	setNumFunctionalUnits( INT_UNIT, 1 );
	setNumFunctionalUnits( FLOAT_UNIT, 1 );
	setNumFunctionalUnits( LDST_UNIT, 1 );
	setNumFunctionalUnits( BRANCH_UNIT, 1 );
	setNumFunctionalUnits( PSEUDO_UNIT, INFINITY );
	setNumFunctionalUnits( SFX_UNIT, 1 );
	setNumFunctionalUnits( CFX_UNIT, 1 );

	/* Instruction Types */
	setInstructionUnitMapping( INTEGER_INSTR, INT_UNIT );
	setInstructionUnitMapping( FLOAT_INSTR, FLOAT_UNIT );
	setInstructionUnitMapping( LDST_INSTR, LDST_UNIT );
	setInstructionUnitMapping( BRANCH_INSTR, BRANCH_UNIT );
	setInstructionUnitMapping( PSEUDO_INSTR, PSEUDO_UNIT );
	setInstructionUnitMapping( SIMPLEINT_INSTR, SFX_UNIT );
	setInstructionUnitMapping( COMPLEXINT_INSTR, CFX_UNIT );

	/* Instruction Preference Order */
	setInstructionTypePreference( 0, INSTR_DONT_CARE );
	setInstructionTypePreference( 1, INSTR_DONT_CARE );
	setInstructionTypePreference( 2, INSTR_DONT_CARE );
	setInstructionTypePreference( 3, INSTR_DONT_CARE );
	setInstructionTypePreference( 4, INSTR_DONT_CARE );
	setInstructionTypePreference( 5, INSTR_DONT_CARE );
	setInstructionTypePreference( 6, INSTR_DONT_CARE );

	/* Set base path to data */
	strcpy(baseDataPath, "");

	/* Set CSP timeout levels */
	setCSPTimeout( TIME_LIMIT_FIRST, 5.0 );
	setCSPTimeout( TIME_LIMIT_SECOND, 15.0 );
	setCSPTimeout( TIME_LIMIT_TOTAL, 600.0 );

	/* Set maximum CSP region and restart sizes */
	setMaxRegionSize( 25 );
	setMaxRestartSize( 300 );

	/* Heuristics */
	setLSHeuristic( HEURISTIC_CRITICAL_PATH );
	setOptHeuristic( HEURISTIC_BEST );

	/* Speculation */
	speculation = TRUE;
}

BOOL isSchedulingSuperblocks()
{
	return schedulingSuperblocks;
}

BOOL shouldRunLS()
{
	return requestedLS;
}

BOOL shouldRunOptimal()
{
	return requestedOptimal;
}

BOOL shouldPrintSchedule()
{
	return requestedPrintSchedule;
}

BOOL isLSForward()
{
	return lsForward;
}

char *
getBaseDataPath()
{
	return baseDataPath;
}


BOOL speculate()
{
	return speculation;
}
