
#include "global.h"
#include "processor.h"
#include "list.h"
#include "dag.h"
#include "csp.h"
#include "solve.h"
#include "timer.h"
#include "ls.h"
#include "heuristic.h"
#include "instr.h"

BOOL propagateBinaryGCC( CSP_type *C, int type, int width, BOOL checkOnly );
static int solveArticulationNodes( CSP_type *C );
static int enumerateAndTest( CSP_type *C );

/*
 *  Choose the next variable to be branched on.
 */
static int
getNextVariable( CSP_type *C, int currentLevel )
{
    int    i, domainSize, minDomainSize, minVariable, maxChanges;

    maxChanges = 0;
    if( C->propagationLevel > 0 ) {
	for( i = 1; i <= C->n; i++ ) if( C->upper[i] > C->lower[i] ) {
	    if( maxChanges < C->changev[i] ) {
		maxChanges = C->changev[i];
	    }
	}
    }

    minDomainSize = INFINITY;
    minVariable = 0;
    for( i = 1; i <= C->n; i++ ) if( C->upper[i] > C->lower[i] ) {

	domainSize = C->upper[i] - C->lower[i] + 1;

	/*
	 *  0.80 appears to be a good value.
	 *  If propagationLevel == 0, all variables will be considered.
	 *  If propagationLevel >  0, only variables near enough to max
	 *      changes will be considered (there will be at least one).
	 */
	if( C->changev[i] >= (0.8*maxChanges) ) {

	    if( (domainSize < minDomainSize) ) {
		minDomainSize = domainSize;
		minVariable = i;
	    }
	    else
	    if( (domainSize == minDomainSize) &&
		(C->nPreds[i] > C->nPreds[minVariable]) ) {
		minVariable = i;
	    }
	    else
	    if( (domainSize == minDomainSize) &&
		(C->nPreds[i] == C->nPreds[minVariable]) &&
		(C->nSuccs[i] > C->nSuccs[minVariable]) ) {
		minVariable = i;
	    }
	}
    }

    return( minVariable );
}

/*
 *  Perform singleton consistency to max depth.
 */
static int
singletonConsistency( CSP_type *C, int currentLevel, int depth, int max_depth )
{
    int  a, b, i;
    BOOL consistent, changed;
int iter;

    saveBounds( C, currentLevel+depth );

    for( i = 1; i <= C->n; i++ ) {
	C->changev[i] = 0;
    }

    for( iter = 0; iter <= 1; iter++ )
    for( i = 1; i <= C->n; i++ ) if( (iter==0 && C->nPreds[i]>1)
					|| (iter==1 && C->nPreds[i]<=1) ){

	if( C->lower[i] == C->upper[i] ) {
	    continue;
	}

	changed = FALSE;

#ifdef DEBUG
	if( depth==0 )
	    printf("sc: variable %d: [%d, %d], currentLevel = %d\n",
			i,C->lower[i],C->upper[i],currentLevel);
#endif

	if( limit_exceeded( C ) ) {
	    return( TIMED_OUT );
	}

	a = C->lower[i];
	b = C->upper[i];

	consistent = FALSE;
	for( ; a <= b && !consistent; a++ ) {
	    C->lower[i] = a;
	    C->upper[i] = a;
	    consistent = propagate( C, i, currentLevel );
	    C->changev[i] = C->changes;
	    if( consistent && (depth < max_depth) ) {
		consistent = singletonConsistency( C, currentLevel, depth+1, max_depth );
	    }
	    restoreBounds( C, currentLevel+depth );
	    if( consistent && (C->lower[i] != a) ) {
	        C->lower[i] = a;
		changed = TRUE;
	        saveBounds( C, currentLevel+depth );
	    }
	}
	if( !consistent || (C->lower[i] > C->upper[i]) ) {
	    return( NO_SOLUTION );
	}

	consistent = FALSE;
	for( ; b >= a && !consistent; b-- ) {
	    C->lower[i] = b;
	    C->upper[i] = b;
	    consistent = propagate( C, i, currentLevel );
	    if( consistent && (depth < max_depth) ) {
		consistent = singletonConsistency( C, currentLevel, depth+1, max_depth );
	    }
	    restoreBounds( C, currentLevel+depth );
	    if( consistent && (C->upper[i] != b) ) {
	        C->upper[i] = b;
		changed = TRUE;
	        saveBounds( C, currentLevel+depth );
	    }
	}

	if( !consistent && (a > b) && (C->upper[i] > C->lower[i]) ) {
	    C->upper[i] = C->lower[i];
	    changed = TRUE;
	    saveBounds( C, currentLevel+depth );
	}
	if( C->lower[i] > C->upper[i] ) {
	    return( NO_SOLUTION );
	}
#ifdef DEBUG
	if( depth==0 )
	    printf("    variable %d: [%d, %d], currentLevel = %d\n",
		i,C->lower[i],C->upper[i],currentLevel);
#endif
	if( changed ) {
	    if( !propagate( C, i, currentLevel ) ) {
		return( NO_SOLUTION );
	    }
	    saveBounds( C, currentLevel+depth );
	}
    }

    return( SOLUTION );
}

/*
 *  Prune the upper bounds of the domains of the cost
 *  variables using the cost function and a known upper
 *  bound on the optimal value of the cost function.
 */
static void
pruneCostVariablesUpper( CSP_type *C )
{
    int  a, b, i, j, ub;
    BOOL consistent;

    for( j = 0; j <= C->nSideExits; j++ ) {
	ub = getLengthUpperBound( C, j );
	if( C->upper[C->costVars[j]] > ub ) {
	    C->upper[C->costVars[j]] = ub;
	    consistent = propagate( C, C->costVars[j], 1 );
	    if( !consistent ) {
		return;
	    }
	}
    }

    saveBounds( C, 1 );

    for( j = 0; j < C->nSideExits; j++ ) {

	i = C->costVars[j];
	if( C->lower[i] == C->upper[i] ) {
	    continue;
	}

	a = C->lower[i];
	b = C->upper[i];

	consistent = FALSE;
	for( ; b >= a && !consistent; b-- ) {
	    C->lower[i] = b;
	    C->upper[i] = b;
	    consistent = propagate( C, i, 1 );
	    if( consistent ) {
		ub = getLengthUpperBound( C, j );
		if( ub < b ) {
		    consistent = FALSE;
		}
	    }
	    restoreBounds( C, 1 );
	    if( consistent && (C->upper[i] != b) ) {
	        C->upper[i] = b;
	        saveBounds( C, 1 );
	    }
	}
    }
}

/*
 *  Determine whether any values can be pruned from the
 *  domains of the cost variables using singleton consistency.
 *  Only looks at the lower bounds of the domains.
 */
static void
pruneCostVariablesLower( CSP_type *C )
{
    int  a, b, i, j, lb, ub;
    BOOL consistent;

    for( j = 0; j <= C->nSideExits; j++ ) {
	lb = getLengthLowerBound( C, j );
	if( C->lower[C->costVars[j]] < lb ) {
	    C->lower[C->costVars[j]] = lb;
	    consistent = propagate( C, C->costVars[j], 1 );
	    if( !consistent ) {
		return;
	    }
	}
    }

    saveBounds( C, 1 );

    for( j = 0; j < C->nSideExits; j++ ) {

	i = C->costVars[j];
	if( C->lower[i] == C->upper[i] ) {
	    continue;
	}

	a = C->lower[i];
	b = C->upper[i];

	consistent = FALSE;
	for( ; a <= b && !consistent; a++ ) {
	    C->lower[i] = a;
	    C->upper[i] = a;
	    consistent = propagate( C, i, 1 );
	    if( consistent ) {
		lb = getLengthLowerBound( C, j );
		if( a < lb ) {
		    consistent = FALSE;
		}
		ub = getLengthUpperBound( C, j );
		if( ub < a ) {
		    consistent = FALSE;
		}
	    }
	    restoreBounds( C, 1 );
	    if( consistent && (C->lower[i] != a) ) {
	        C->lower[i] = a;
	        saveBounds( C, 1 );
	    }
	}
    }
}

/*
 *  Determine whether any values can be pruned from the
 *  domains of the cost variables.
 */
static int
pruneCostVariables( CSP_type *C )
{
    int  consistent, status, cost;

    consistent = propagate( C, C->costVars[C->nSideExits], 0 );
    if( !consistent ) {
	return( FALSE );
    }

    pruneCostVariablesLower( C );
    pruneCostVariablesUpper( C );

    cost = getCost( C );
    if( C->lbCost < cost ) {
	C->lbCost = cost;
	if( C->lbCost == C->ubCost ) {
	    return( TRUE );
	}
	if( C->lbCost > C->ubCost ) {
	    return( FALSE );
	}
    }

    status = singletonConsistency( C, 1, 0, /*max_depth=*/0 );
    if( status == NO_SOLUTION ) {
	return( FALSE );
    }

    pruneCostVariablesLower( C );
    pruneCostVariablesUpper( C );

    cost = getCost( C );
    if( C->lbCost < cost ) {
	C->lbCost = cost;
	if( C->lbCost == C->ubCost ) {
	    return( TRUE );
	}
	if( C->lbCost > C->ubCost ) {
	    return( FALSE );
	}
    }

    status = singletonConsistency( C, 1, 0, /*max_depth=*/1 );
    if( status == NO_SOLUTION ) {
	return( FALSE );
    }

    pruneCostVariablesLower( C );
    pruneCostVariablesUpper( C );

    cost = getCost( C );
    if( C->lbCost < cost ) {
	C->lbCost = cost;
	if( C->lbCost == C->ubCost ) {
	    return( TRUE );
	}
	if( C->lbCost > C->ubCost ) {
	    return( FALSE );
	}
    }

    return( TRUE );
}

/*
 *  Interface to singletonConsistency routine.
 *  Returns SOLUTION if domains are non-empty after probing.
 *  Returns NO_SOLUTION if domains are empty after probing.
 *  Returns TIMED_OUT if time expired before probing was complete.
 */
static int
probe( CSP_type *C, int currentLevel )
{
    return( singletonConsistency( C, currentLevel,
		/*depth=*/0, /*max_depth*/C->propagationLevel-1 ) );
}


/*
 *  Solve the CSP by backtracking search interleaved with
 *  constraint propagation and chronological backtracking.
 *  Returns SOLUTION if a solution has been found for the CSP.
 *  Returns NO_SOLUTION if CSP has been proven inconsistent.
 *  Returns TIMED_OUT if time expired before backtracking was complete.
 */
static int
backtrack( CSP_type *C, int currentLevel )
{
    int  a, b, currentVar, status;

    saveBounds( C, currentLevel );

    if( C->propagationLevel > 0 ) {
		// Pruning has been disabled since it doesn't properly account for
		// issue width or pipelining constraints
		//safePrune( C );
		status = probe( C, currentLevel );
		if( status != SOLUTION ) {
			return( status );
		}
    }

    currentVar = getNextVariable( C, currentLevel );

    if( currentVar == 0 ) {
        return( SOLUTION );
    }
#ifdef DEBUG
    if( C->propagationLevel > 0 )
	printf("%d: next variable at level %d is %d, [%d, %d], %d\n",
		C->n,currentLevel,currentVar,
		C->lower[currentVar],
		C->upper[currentVar],
		C->varType[currentVar]);
#endif

    a = C->lower[currentVar];
    b = C->upper[currentVar];

    for( ; a <= b; a++ ) {

	C->lower[currentVar] = a;
	C->upper[currentVar] = a;
#ifdef DEBUG
    if( C->propagationLevel > 0 )
	printf("\t%d <-- %d, remaining: [%d,%d], level: %d\n",currentVar,a,a+1,b,currentLevel);
#endif

	C->visits++;
//TEMPORARY && (C->backtracks > 10)
	if( limit_exceeded( C ) && (C->backtracks > 10) ) {
	    return( TIMED_OUT );
	}

        if( propagate( C, currentVar, currentLevel ) ) {
            if( (status = backtrack( C, currentLevel+1 )) ) {
                return( status );
            }
        }

        restoreBounds( C, currentLevel );
    }

    C->backtracks++;

    return( NO_SOLUTION );
}

/*
 *  Interface to backtracking routine.
 *  Called many times to solve sub-CSPs and again to solve entire CSPs.
 *  If this code is changed, be sure it works for all call sites.
 *  Returns PROVED_OPTIMAL, length is length of optimal length solution.
 *  Returns SOLUTION, length is length of optimal length solution.
 *  Returns TIMED_OUT, length is lower bound on length of optimal solution.
 */
int
solve( CSP_type *C, int upperBound, int *length )
{
    int	 status;
	BOOL error = FALSE;

#ifdef USE_EXECUTION_SERIALIZATION
	int i;
#endif

    saveBounds( C, 0 );

    while( 1 ) {
	*length = C->m;

	/*
	 *  If the number of cycles needed to schedule the instructions
	 *  is equal to the known upper bound on the schedule, we're done.
	 */
	if( C->m == upperBound ) {
	    return( PROVED_OPTIMAL );
	}

#ifdef DEBUG
	printf("%d: attempting schedule length %d\n",C->n,C->m);
#endif

	// Start with an initial binary GCC check. This is added since for
	// some really small blocks, the initial bounds force a serial conflict,
	// but propagateBinaryGCC is never called for relevant variables.
#ifdef USE_EXECUTION_SERIALIZATION
	error = FALSE;
	for( i = 0; i < N_FUNCTIONAL_UNIT_TYPES; i++ ) {
		if( !propagateBinaryGCC( C, i, nFunctionalUnits( i ), TRUE ) ) {
			error = TRUE;
			break;
		}
	}

#endif

	// Propagate as normal
	if( error == FALSE && propagate( C, 1, /*currentLevel=*/0 ) ) {
		status = backtrack( C, 1 );
	}
	else					    status = NO_SOLUTION;

	if( status == SOLUTION ) {
	    return( SOLUTION );
	}

	if( status == NO_SOLUTION ) {
	    restoreBounds( C, 0 );
	    incrementUpperBounds( C );
	    saveBounds( C, 0 );
	}

	if( status == TIMED_OUT ) {

	    if( !C->restartsAllowed ) {
		return( TIMED_OUT );
	    }
	    if( (C->n <= getMaxRestartSize()) && (C->propagationLevel == 0) ) {
		set_relative_CPU_limit( C, getCSPTimeout( TIME_LIMIT_SECOND ) );
		C->propagationLevel++;
		restoreBounds( C, 0 );
	        status = NO_SOLUTION;
	    }
	    else if( (C->n <= getMaxRestartSize()) && (C->propagationLevel == 1) ) {
		set_absolute_CPU_limit( C, getCSPTimeout( TIME_LIMIT_TOTAL ) );
		C->propagationLevel++;
		restoreBounds( C, 0 );
	        status = NO_SOLUTION;
	    }
	    else if( (C->n >  getMaxRestartSize()) && (C->propagationLevel == 0) ) {
		set_absolute_CPU_limit( C, getCSPTimeout( TIME_LIMIT_TOTAL ) );
		C->propagationLevel++;
		restoreBounds( C, 0 );
	        status = NO_SOLUTION;
	    }
	    else {
		return( TIMED_OUT );
	    }

	}

    }
}

void optimalScheduler( DAG_type *DAG, int bestSchedule[], int* foundOptimal, int* finalCost, int* finalLength, double *costWithFreq )
{
    CSP_type    *C;
    int         *schedule;
    int		lbLength; // lower bound on length of optimal schedule
    int		ubLength = INFINITY; // upper bound on length of optimal schedule
    int		i, type, status, cost, j;
	int save_m, lb, ub, consistent;
	int currentLength, currentCost;
    int secondAttempt;
    double costF;

#ifdef USE_PIPELINING
	int pos = 0;

#ifdef USE_EXECUTION_SERIALIZATION
	// Need a count of extra nodes due to serial constraints
	for( i = 1; i <= DAG->nInstructions; i++ ) {
		if( isExecution( DAG->instructionIndex[i - 1] ) ) {
			++pos;
		}
	}
#endif

	// If using pipelining, need to create dummy nodes
	C = createCSP( DAG->nInstructions + DAG->nDummyNodes + (pos * getMaxUnitsOfAnyType()) + 1 );
	// Don't want to use dummy nodes, in general, as real nodes
	C->n = DAG->nInstructions;
	C->nDummyNodes = DAG->nDummyNodes;
#else
	C = createCSP( DAG->nInstructions + 1 );
	C->n = DAG->nInstructions;
#endif

    for( i = 1; i <= C->n; i++ ) {
		C->varType[i] = DAG->instructionType[i-1];
#ifdef USE_PIPELINING
		C->varDuration[i] = instructionTimes[DAG->instructionIndex[i - 1]][0];
#endif

#ifdef USE_EXECUTION_SERIALIZATION
		C->varIsSerial[i] = isExecution( DAG->instructionIndex[i - 1] );
#endif
    }
    for( type = 0; type < N_INSTRUCTION_TYPES; type++ ) {
	C->nOps[type] = 0;
    }
    for( i = 1; i <= C->n; i++ ) {
	C->nOps[C->varType[i]]++;
	C->changev[i] = 0;
    }

#ifdef USE_PIPELINING
	// Need to set data for the dummy nodes
	pos = C->n + 1;
	for( i = 1; i < C->n; i++ ) {
		if( C->varDuration[i] > 1) {
			for( j = 1; j < C->varDuration[i]; j++ ) {
				C->varDuration[pos] = 1;
				C->varType[pos] = C->varType[i];
				pos++;
			}
		}
	}
#endif

    schedule = (int *)calloc( (size_t)DAG->nInstructions, sizeof(int) );

    setCostFunction( C, DAG );
#ifdef TRACK
    printCostFunction( C );
#endif
    setConstraints( C, DAG );
    setBounds( C );

    /*
     *  Now that the bounds have been set, determine a lower
     *  bound on the cost of an optimal cost schedule using the
     *  lower bounds on the cost variables.
     */
    C->lbCost = getCost( C );

    /*
     *  Find a feasible schedule using a list scheduler.
     *  The length of the feasible schedule is an upper bound on
     *  the length of an optimal length schedule (used by solve).
     *  The cost of the feasible schedule is an upper bound on the
     *  cost of an optimal cost schedule.
     */
    C->ubCost = INFINITY;
    if ( getOptHeuristic() == HEURISTIC_BEST ) {
        // If best, try for each heuristic
        for ( i = 0; i < HEURISTIC_NUM_HEURISTICS; i++ ) {
            if ( i != HEURISTIC_BEST ) {
                currentCost = 0;
                currentLength = 0;
                costF = 0;
                listScheduler( DAG, i, schedule, &currentCost, &currentLength, &costF );
                if( C->ubCost > currentCost ) {
                    for( j = 1; j <= C->n; j++ ) {
                        bestSchedule[j-1] = schedule[j-1];
                    }
                }
                ubLength = min( ubLength, currentLength );
                C->ubCost = min( C->ubCost, currentCost );
            }
        }
    }
    else {
        listScheduler( DAG, getOptHeuristic(), bestSchedule, &C->ubCost, &ubLength, &costF );
    }

    if( C->lbCost == C->ubCost ) {
		goto END;
    }

#ifdef TRACK
printf("After heuristics: %d <= c <= %d\n",C->lbCost,C->ubCost);
fflush(stdout);
#endif

    save_m = C->m;
    saveBounds( C, 0 );
    saveBounds( C, C->n );
    status = solveArticulationNodes( C );
    if( status == SOLUTION ) {
        verifySolution( C );
	cost = getCost( C );
	if( cost < C->ubCost ) {
	    C->ubCost = cost;
            for( i = 1; i <= C->n; i++ ) {
	        bestSchedule[i-1] = C->lower[i];
            }
        }
	if( allArticulationNodes( C ) ) {
	    C->lbCost = C->ubCost = cost;
	}
    }
    restoreBounds( C, C->n );
    C->m = save_m;
    saveBounds( C, 0 );
#ifdef TRACK
printf("After solveArticulationNodes: %d <= c <= %d\n",C->lbCost,C->ubCost);
fflush(stdout);
#endif

    if( C->lbCost == C->ubCost ) {
	goto END;
    }

    /*
     *  Determine an optimal length schedule.
     *  The length of the optimal length schedule is a lower
     *  bound on the length of an optimal cost schedule.
     *  Options to test here: restartsAllowed = FALSE, ubLength = INFINITY.
     */
    set_relative_CPU_limit( C, getCSPTimeout( TIME_LIMIT_FIRST ) );
    C->propagationLevel = 0;
    C->restartsAllowed = FALSE;
    secondAttempt = FALSE;
next_attempt:
    status = solve( C, INFINITY, &lbLength );
    //status = solve( C, ubLength, &lbLength );

    /*
     *  Check whether optimal length schedule is of lower cost.
     *  If it is of lower cost, update bestSchedule.
     */
    if( status == SOLUTION ) {
        verifySolution( C );
	cost = getCost( C );
	if( cost < C->ubCost ) {
	    C->ubCost = cost;
            for( i = 1; i <= C->n; i++ ) {
	        bestSchedule[i-1] = C->lower[i];
            }
        }
    }
#ifdef TRACK
printf("After solve: %d <= c <= %d\n",C->lbCost,C->ubCost);
fflush(stdout);
#endif

    if( C->lbCost == C->ubCost ) {
	goto END;
    }

    /*
     *  The upper bound on the cost of an optimal cost schedule
     *  can be used to determine upper bounds on the domains of
     *  the cost variables, including an upper bound on the
     *  length of the optimal cost schedule.
     */
    restoreBounds( C, 0 );
    ubLength = getLengthUpperBound( C, C->nSideExits );

    /*
     *  Can the bounds be improved?
     *
     *  Make sure that any pruning of values is valid
     *  for all feasible schedules, and then attempt to
     *  prune values from the domains of the cost variables.
     */
    addToUpperBounds( C, ubLength - lbLength );
    set_relative_CPU_limit( C, getCSPTimeout( TIME_LIMIT_SECOND ) );
    consistent = pruneCostVariables( C );
    if( !consistent ) {
	//fprintf( stderr, ":error pruneCostVariables inconsistent\n" );
	//exit( 6 );
	restoreBounds( C, C->n );
	C->m = save_m;
	saveBounds( C, 0 );
	if( !secondAttempt ) {
	    set_relative_CPU_limit( C, getCSPTimeout( TIME_LIMIT_FIRST ) );
	    C->propagationLevel = 0;
	    C->restartsAllowed = TRUE;
	    secondAttempt = TRUE;
	    goto next_attempt;
	}
    }
    saveBounds( C, 0 );
    cost = getCost( C );
    if( C->lbCost < cost ) {
	C->lbCost = cost;
    }

#ifdef TRACK
printf("After pruneCostVariables: %d <= c <= %d\n",C->lbCost,C->ubCost);
fflush(stdout);
#endif
    if( C->lbCost == C->ubCost ) {
	goto END;
    }

    /*
     *  General case.
     *
     *  Given the domains of the cost variables and lower and
     *  upper bounds on the cost functions:
     *  - enumerate the assignments to the cost variables
     *    which satisfy the bounds on the cost function
     *  - sort the assignments by increasing cost
     *  - step through the assignments until one is found that
     *    can be extended to a solution to the entire CSP
     */
    status = enumerateAndTest( C );
#ifdef TRACK
printf("After first call to enumerateAndTest: %d <= c <= %d\n",C->lbCost,C->ubCost);
fflush(stdout);
#endif

    if( status == OUT_OF_SPACE ) {
	lb = C->lbCost;
	ub = C->ubCost;
	status = NO_SOLUTION;
	for( cost = lb; cost < ub && status == NO_SOLUTION; cost++ ) {
	    restoreBounds( C, 0 );
	    C->lbCost = cost;
	    C->ubCost = cost;
	    consistent = pruneCostVariables( C );
	    if( consistent ) {
		status = enumerateAndTest( C );
	    }
	}
	if( (status == OUT_OF_SPACE) || (status == TIMED_OUT) ) {
	    /*
	     *  Restore the saved upper bound on the cost if we ran
	     *  out of space or time. The lower bound on the cost is
	     *  still valid.
	     */
	    C->ubCost = ub;
	}
	else
	if( status == NO_SOLUTION ) {
	    /*
	     *  If no solution was found of cost c, where lb <= c < ub,
	     *  then the solution we have of cost ub has been proven optimal.
	     */
	    C->lbCost = C->ubCost = ub;
	}
	/*
	 *  The fall through case is (status == SOLUTION), in which case
         *  the bounds remain as they were set inside the loop.
	 *  The case (status == PROVED_OPTIMAL) cannot happen here.
	 */
    }

    if( status == SOLUTION ) {
        verifySolution( C );
        for( i = 1; i <= C->n; i++ ) {
	    bestSchedule[i-1] = C->lower[i];
        }
    }

END:
#ifdef DEBUG
    printf("%s: n: %d, lc: %d, uc: %d, dc: %d, t: %0.2f\n",
	DAG->blockID, DAG->nInstructions,
	C->lbCost, C->ubCost, C->ubCost - C->lbCost, elapsed_time_CSP( C ) );
    fflush( stdout );
#endif

    *foundOptimal = (C->ubCost == C->lbCost);
    *finalCost = C->ubCost;

    costF = 0.0;
    for( i = 0; i <= C->nSideExits; i++ ) {
		costF += (double)C->costCoeffs[i] *
			     (double)bestSchedule[C->costVars[i]-1] *
			     C->costFreqs[i];
    }
    *costWithFreq = costF;

    destroyCSP( C );
    free( schedule );
}

static int
solveArticulationNodes( CSP_type *C )
{
    BOOL     *inRegion;
    int      i, x, w, distance, status;
	int      distanceOutOfBounds;

    inRegion = (BOOL *)calloc( (size_t)C->n+1, sizeof(BOOL) );

	for( i = 0; i < C->nSideExits; i++ ) {
		x = C->costVars[i];

		/*
		*  Mark the nodes w that are necessarily between
		*  1 and x (i.e., on a path from 1 to x).
		*/
		for( w = 1; w <= C->n; w++ ) {
			inRegion[w] = FALSE;
		}
		inRegion[1] = TRUE;
		inRegion[x] = TRUE;
		for( w = 1; w <= C->n; w++ ) {
			if( (C->cp[1][w] >= 0) && (C->cp[w][x] >= 0) ) {
				inRegion[w] = TRUE;
			}
		}
		status = solveRegion( C, 1, x, inRegion, &distance );
		if( status == TIMED_OUT ) {
			goto END;
		}

		// See if the side exit exceeds its bounds
		distanceOutOfBounds = 0;
		if( (distance + 1) > C->upper[x] ) {
			distanceOutOfBounds = (distance + 1) - C->upper[x];
		}
		
		// Doesn't matter if the region can be solved optimally with
		// a size of distance or not; if the side exit is given a lower
		// bound that is lower than its previous lower bound, constraints
		// will be violated. So, we adjust distance so that nothing changes
		if( (distance + 1) < C->lower[x] ) {
			distance = C->lower[x] - 1;
		}

		C->lower[x] = C->upper[x] = distance + 1;
		C->fixed[x] = TRUE;

		// If the side exit exceeded its bounds, need to adjust bounds
		// for any descendants of the side exit
		if( distanceOutOfBounds > 0 ) {
			for( w = 1; w <= C->n; w++ ) {
				// If w is a descendant of x, add the distance by which
				// x exceeded its bounds to w's upper bound
				if( w != x && C->cp[x][w] >= 0 ) {
					C->upper[w] += distanceOutOfBounds;
				}
			}

			// Fix up bounds for the final node
			C->lower[C->n] = C->upper[C->n];
			C->m = C->upper[C->n];

			// The above change only set upper bounds for descendant nodes.
			// Need to propagate changes so that lower bounds are tightened.
			//propagate( C, x, 0 );
			if( !propagate( C, x, 0 ) ) {
			    status = NO_SOLUTION;
			    goto END;
			}
		}
	}

    /*
     *  Allow restarts only if all the side exits are articulation
     *  nodes. In this case solving the CSP gives the optimal
     *  solution, so the time is well spent.
     */
    set_relative_CPU_limit( C, getCSPTimeout( TIME_LIMIT_FIRST ) );
    C->propagationLevel = 0;
    C->restartsAllowed = allArticulationNodes( C );
    status = solve( C, INFINITY, &distance );

END:
    for( i = 0; i < C->nSideExits; i++ ) {
        x = C->costVars[i];
	C->fixed[x] = FALSE;
    }
    free( inRegion);

    return( status );
}

// TEMPORARY: For printing assignments
static void
print_Info( CSP_type *C, int **all, int n_assignments, int *map, int *cost )
{
    int i, j;
    printf("\tnSideExits = %d\n", C->nSideExits);
    printf("\tC->lbCost = %d\n", C->lbCost);
    printf("\tC->ubCost = %d\n", C->ubCost);
    printCostFunction( C );
    for( i = 0; i <= C->nSideExits; i++ ) {
      printf("\tx_%d = [%d, %d]\n",
        C->costVars[i],
        C->lower[C->costVars[i]],
        C->upper[C->costVars[i]]);
    }
    printf("n_assignments = %d\n", n_assignments);
    for( i = 0; i < n_assignments; i++ ) {
        for( j = 0; j <= C->nSideExits; j++ ) {
	    printf("%d ", all[map[i]][j]);
        }
        printf(": %d\n", cost[map[i]]);
    }
}
// END TEMPORARY: For printing assignments

static void
sortmin( int map[], int cost[], int n )
{
    int  i, t, current;
    BOOL sorted;

    current = n-1;
    sorted = FALSE;
    while( !sorted ) {
	sorted = TRUE;
	for( i = 0; i < current; i++ ) {
	    if( cost[map[i]] > cost[map[i+1]] ) {
		t = map[i];
		map[i] = map[i+1];
		map[i+1] = t;
		sorted = FALSE;
	    }
	}
	current--;
    }
}

/*
 *  Given the domains of the cost variables and lower and
 *  upper bounds on the cost functions:
 *  - enumerate the assignments to the cost variables
 *    which satisfy the bounds on the cost function
 *  - sort the assignments by increasing cost
 *  - step through the assignments until one is found that
 *    can be extended to a solution to the entire CSP
 *  Returns PROVED_OPTIMAL, lbCost, ubCost are cost of optimal cost solution.
 *  Returns SOLUTION, lbCost, ubCost are cost of optimal cost solution.
 *  Returns TIMED_OUT, lbCost, ubCost are bounds on cost of optimal solution.
 */
static int
enumerateAndTest( CSP_type *C )
{
    int i, j, consistent, status = NO_SOLUTION, n_assignments = 0, value;
    int *partial, **all, *map, *cost;
    int equalCostBounds;

    equalCostBounds = (C->lbCost == C->ubCost);

    all = (int **)calloc( (size_t)MAX_NUMBER_ASSIGNS, sizeof(int *) );
    partial = (int *)calloc( (size_t)C->nSideExits+1, sizeof(int) );

    /*
     *  Enumerate the assignments to the cost variables
     *  which satisfy the bounds on the cost function.
     */
    enumerateCostVariables( C, all, &n_assignments, partial, 0 );
    if( n_assignments == 0 ) {
#ifdef TRACK
printf("enumerateAndTest: n_assignments == 0\n");
#endif
	status = NO_SOLUTION;
	goto END;
    }
    if( n_assignments >= MAX_NUMBER_ASSIGNS ) {
#ifdef TRACK
printf("enumerateAndTest: Out of space\n");
#endif
	status = OUT_OF_SPACE;
	goto END;
    }

    /*
     *  Sort the assignments by increasing cost.
     */
    map  = (int *)calloc( (size_t)n_assignments, sizeof(int) );
    cost = (int *)calloc( (size_t)n_assignments, sizeof(int) );
    for( i = 0; i < n_assignments; i++ ) {
	map[i] = i;
	cost[i] = 0;
	for( j = 0; j <= C->nSideExits; j++ ) {
	    cost[i] += C->costCoeffs[j] * all[i][j];
	}
    }

    if( !equalCostBounds ) {
	sortmin( map, cost, n_assignments );
    }
#ifdef TRACK
    print_Info( C, all, n_assignments, map, cost );
#endif

    /*
     *  Step through the assignments until one is found that
     *  can be extended to a solution to the entire CSP.
     */
    status = NO_SOLUTION;
    for( i = 0; (i < n_assignments) && (status == NO_SOLUTION); i++ ) {
#ifdef TRACK
    printf("Track: enumerateAndTest: %d of %d, cost = %d\n",
	i+1, n_assignments, cost[map[i]]);
    fflush(stdout);
#endif
if( getGlobalStopwatchElapsedTime() > getCSPTimeout( TIME_LIMIT_TOTAL ) ) {
    return( TIMED_OUT );
}

	C->lbCost = cost[map[i]];
	if( (C->lbCost == C->ubCost) && !equalCostBounds ) {
	    status = PROVED_OPTIMAL;
	}
	else {
	    restoreBounds( C, 0 );
	    /*
	     *  Fix cost variables.
	     */
	    for( j = 1; j <= C->n; j++ ) {
		C->fixed[j] = FALSE;
	    }
	    consistent = TRUE;
	    for( j = 0; j <= C->nSideExits && consistent; j++ ) {
		value = all[map[i]][j];
		if( (value < C->lower[C->costVars[j]]) ||
		    (value > C->upper[C->costVars[j]]) ) {
		    consistent = FALSE;
		}
		else {
		    C->lower[C->costVars[j]] = value;
		    C->upper[C->costVars[j]] = value;
		    C->fixed[C->costVars[j]] = TRUE;
		    consistent = propagate( C, C->costVars[j], 0 );
		}
	    }

	    /*
	     *  Solve.
	     */
	    if( consistent ) {
		set_relative_CPU_limit( C, getCSPTimeout( TIME_LIMIT_FIRST ) );
		C->propagationLevel = 0;
		saveBounds( C, 1 );
		status = backtrack( C, 1 );
		if( status == TIMED_OUT ) {
		    set_relative_CPU_limit( C, getCSPTimeout( TIME_LIMIT_SECOND ) );
		    C->propagationLevel = 1;
		    restoreBounds( C, 1 );
		    status = backtrack( C, 1 );
		    if( status == TIMED_OUT ) {
			if( elapsed_time_CSP( C ) > getCSPTimeout( TIME_LIMIT_TOTAL ) ) {
			    return( TIMED_OUT );
			}
			else {
			    set_absolute_CPU_limit( C, getCSPTimeout( TIME_LIMIT_TOTAL ) );
			    if( C->n <= getMaxRestartSize() ) {
				C->propagationLevel++;
			    }
			    restoreBounds( C, 1 );
			    status = backtrack( C, 1 );
			    if( status == TIMED_OUT ) {
				return( TIMED_OUT );
			    }
			}
		    }
		}
	    }
	    if( status == SOLUTION ) {
		C->ubCost = cost[map[i]];
	    }
	}
    }

    free( map );
    free( cost );

END:
    for( i = 0; i < n_assignments; i++ ) {
	free( all[i] );
    }
    free( all );
    free( partial );

    return( status );
}

