
/* =================================================================
 *  Routines for getting the amount of CPU time used.
 */

#include <stdio.h>
#include "global.h"
#include "processor.h"
#include "list.h"
#include "dag.h"
#include "csp.h"
#include "timer.h"


#ifndef WIN32 /* Unix-specific timer code */
	#include <sys/time.h>
	#include <sys/resource.h>
#else

#include <windows.h>

struct rusage {
    struct timeval ru_utime, ru_stime;
};

#define RUSAGE_SELF 0

static getrusage(int ignored, register struct rusage *ru)
{
	HANDLE phd;
	FILETIME CreateTime, ExitTime, KernelTime, UserTime;
	SYSTEMTIME SysTime;
	phd = GetCurrentProcess();
	if ( GetProcessTimes(phd, &CreateTime, &ExitTime, &KernelTime, &UserTime) == FALSE) {
		ru->ru_stime.tv_sec  = 0;
		ru->ru_stime.tv_usec = 0;
		ru->ru_utime.tv_sec  = 0;
		ru->ru_utime.tv_usec = 0;
	}
	else {
		(void) FileTimeToSystemTime(&KernelTime, &SysTime);

		ru->ru_stime.tv_sec  = SysTime.wSecond + (60 * SysTime.wMinute) + (60 * 60 * SysTime.wHour);
		ru->ru_stime.tv_usec = SysTime.wMilliseconds * 1000;
		
		(void) FileTimeToSystemTime(&UserTime, &SysTime);

		ru->ru_utime.tv_sec  = SysTime.wSecond + (60 * SysTime.wMinute) + (60 * 60 * SysTime.wHour);
		ru->ru_utime.tv_usec = SysTime.wMilliseconds * 1000;
	}
}

#endif

static struct rusage res;

/*
 *  Start the stopwatch.
 */
void
set_start_time_CSP( CSP_type *C )
{
    getrusage( RUSAGE_SELF, &res );
    C->startTime = (double) res.ru_utime.tv_sec +
                   (double) res.ru_stime.tv_sec +
                   (double) res.ru_utime.tv_usec / 1000000.0 +
                   (double) res.ru_stime.tv_usec / 1000000.0;
}

/*
 *  Return the CPU time used in seconds.
 */
double
elapsed_time_CSP( CSP_type *C )
{
    getrusage( RUSAGE_SELF, &res );
    return( (double) res.ru_utime.tv_sec +
            (double) res.ru_stime.tv_sec +
            (double) res.ru_utime.tv_usec / 1000000.0 +
            (double) res.ru_stime.tv_usec / 1000000.0
            - C->startTime );
}


/* =================================================================
 *  Routines for setting and testing limits on the amount of CPU
 *  time used.
 */

/*
 *  Set an absolute limit on the CPU time, as
 *  measured from the creation time of the CSP.
 */
void
set_absolute_CPU_limit( CSP_type *C, double time_limit )
{
    C->maxCPUseconds = time_limit;
}

/*
 *  Set a relative limit on the CPU time, as
 *  measured from the current time.
 */
void
set_relative_CPU_limit( CSP_type *C, double time_limit )
{
    C->maxCPUseconds = elapsed_time_CSP( C ) + time_limit;
}

/*
 *  Answer whether the preset limit has been exceeded.
 */
BOOL
limit_exceeded( CSP_type *C )
{
    return( elapsed_time_CSP( C ) > C->maxCPUseconds );
}


/* =================================================================
 *  Routines for global timing
 */

double globalStopwatchTime;

/*
 * Start the global stopwatch
 */
void
startGlobalStopwatch()
{
    getrusage( RUSAGE_SELF, &res );
    globalStopwatchTime =
		           (double) res.ru_utime.tv_sec +
                   (double) res.ru_stime.tv_sec +
                   (double) res.ru_utime.tv_usec / 1000000.0 +
                   (double) res.ru_stime.tv_usec / 1000000.0;
}

/*
 * Get the elapsed time since the stopwatch was started
 */
double
getGlobalStopwatchElapsedTime()
{
    getrusage( RUSAGE_SELF, &res );
    return( (double) res.ru_utime.tv_sec +
            (double) res.ru_stime.tv_sec +
            (double) res.ru_utime.tv_usec / 1000000.0 +
            (double) res.ru_stime.tv_usec / 1000000.0
            - globalStopwatchTime );
}
